# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.witness module

Witness command line interface
"""
import argparse
import logging
import os
import sys

import falcon
from hio.base import doing
from hio.core import http
from hio.core.tcp import serving as tcpServing
from hio.help import decking

from keri import help
from keri import kering
from keri.app import directing, agenting, indirecting, storing, grouping
from keri.app.cli.common import existing
from keri.core import scheming, coring
from keri.peer import exchanging
from keri.vc import walleting, handling, proving
from keri.vdr import verifying, viring


WEB_DIR_PATH = os.path.dirname(
                os.path.abspath(
                    sys.modules.get(__name__).__file__))
STATIC_DIR_PATH = os.path.join(WEB_DIR_PATH, 'static')


d = "Runs KERI Agent controller.\n"
d += "Example:\nagent -t 5621\n"
parser = argparse.ArgumentParser(description=d)
parser.set_defaults(handler=lambda args: launch(args))
parser.add_argument('-T', '--tcp',
                    action='store',
                    default=5621,
                    help="Local port number the HTTP server listens on. Default is 5621.")
parser.add_argument('-a', '--admin-http-port',
                    action='store',
                    default=5623,
                    help="Admin port number the HTTP server listens on. Default is 5623.")
parser.add_argument('-n', '--name',
                    action='store',
                    default="agent",
                    help="Name of controller. Default is agent.")
parser.add_argument('-c', '--controller',
                    action='store',
                    default=None,
                    help="Identifier prefix to accept control messages from.")
parser.add_argument("-I", '--insecure',
                    action='store_true',
                    help="Run admin HTTP server without checking signatures on controlling requests")
parser.add_argument("-p", "--path",
                    action="store",
                    default=STATIC_DIR_PATH,
                    help="Location of the KIWI app bundle for this agent")


def launch(args):
    help.ogler.level = logging.CRITICAL
    help.ogler.reopen(name="keri", temp=True, clear=True)
    logger = help.ogler.getLogger()

    logger.info("\n******* Starting Agent for %s listening: http/%s, tcp/%s "
                ".******\n\n", args.name, args.admin_http_port, args.tcp)
    print("Starting agent", args.name)
    doers = runAgent(controller=args.controller, name=args.name, insecure=args.insecure,
                     tcp=int(args.tcp),
                     adminHttpPort=int(args.admin_http_port), path=args.path)
    try:
        tock = 0.03125
        doist = doing.Doist(limit=0.0, tock=tock, real=True)
        doist.do(doers=doers)
    except kering.ConfigurationError:
        print(f"prefix for {args.name} does not exist, incept must be run first", )


    logger.info("\n******* Ended Agent for %s listening: http/%s, tcp/%s"
                ".******\n\n", args.name, args.admin_http_port, args.tcp)


def runAgent(controller, name="agent", insecure=False, tcp=5621, adminHttpPort=5623, path=STATIC_DIR_PATH):
    """
    Setup and run one agent
    """

    hab, doers = existing.openHabitat(name=name)

    # setup doers
    server = tcpServing.Server(host="", port=tcp)
    tcpServerDoer = tcpServing.ServerDoer(server=server)
    directant = directing.Directant(hab=hab, server=server)

    reger = viring.Registry(name=hab.name, temp=False)
    verifier = verifying.Verifier(hab=hab, name=hab.name, reger=reger)
    wallet = walleting.Wallet(db=verifier.reger, name=name)

    handlers = []

    proofs = decking.Deck()
    issuerCues = decking.Deck()

    jsonSchema = scheming.JSONSchema(resolver=scheming.jsonSchemaCache)
    issueHandler = handling.IssueHandler(hab=hab, verifier=verifier, typ=jsonSchema)
    requestHandler = handling.RequestHandler(wallet=wallet, typ=jsonSchema)
    applyHandler = handling.ApplyHandler(hab=hab, verifier=verifier, name=name, issuerCues=issuerCues)
    proofHandler = handling.ProofHandler(proofs=proofs)

    mbx = storing.Mailboxer(name=hab.name)
    mih = grouping.MultisigInceptHandler(hab=hab, controller=controller, mbx=mbx)
    ish = grouping.MultisigIssueHandler(hab=hab, controller=controller, mbx=mbx)
    meh = grouping.MultisigEventHandler(hab=hab, verifier=verifier)

    handlers.extend([issueHandler, requestHandler, proofHandler, applyHandler, mih, ish, meh])

    exchanger = exchanging.Exchanger(hab=hab, handlers=handlers)

    rep = storing.Respondant(hab=hab, mbx=mbx)
    cues = decking.Deck()
    mbd = indirecting.MailboxDirector(hab=hab,
                                      exc=exchanger,
                                      verifier=verifier,
                                      rep=rep,
                                      topics=["/receipt", "/replay", "/multisig", "/credential", "/delegate"],
                                      cues=cues)
    # configure a kevery
    doers.extend([exchanger, directant, tcpServerDoer, mbd])
    doers.extend(adminInterface(controller=controller,
                                hab=hab,
                                insecure=insecure,
                                proofs=proofs,
                                cues=cues,
                                issuerCues=issuerCues,
                                verifier=verifier,
                                mbx=mbx,
                                mbd=mbd,
                                adminHttpPort=adminHttpPort,
                                path=path))

    return doers


def adminInterface(controller, hab, insecure, proofs, cues, issuerCues, mbx, mbd, verifier, adminHttpPort=5623,
                   path=STATIC_DIR_PATH):
    app = falcon.App(middleware=falcon.CORSMiddleware(
        allow_origins='*', allow_credentials='*', expose_headers=['cesr-attachment', 'cesr-date', 'content-type']))
    print("creating static sink for", path)
    sink = http.serving.StaticSink(staticDirPath=path)
    app.add_sink(sink, prefix=sink.DefaultStaticSinkBasePath)

    rep = storing.Respondant(hab=hab, mbx=mbx)

    httpHandler = indirecting.HttpMessageHandler(hab=hab, app=app, rep=rep)
    gdoer = grouping.MultiSigGroupDoer(hab=hab, ims=mbd.ims)

    kiwiServer = agenting.KiwiServer(hab=hab, controller=controller, verifier=verifier, gdoer=gdoer.msgs, app=app,
                                     rep=rep, issuerCues=issuerCues, insecure=insecure)

    mbxer = storing.MailboxServer(app=app, hab=hab, mbx=mbx)
    wiq = agenting.WitnessInquisitor(hab=hab)

    proofHandler = AdminProofHandler(hab=hab, controller=controller, mbx=mbx, verifier=verifier, wiq=wiq, proofs=proofs)
    cueHandler = AdminCueHandler(hab=hab, controller=controller, mbx=mbx, cues=cues)
    server = http.Server(port=adminHttpPort, app=app)
    httpServerDoer = http.ServerDoer(server=server)

    doers = [httpServerDoer, httpHandler, rep, mbxer, wiq, proofHandler, cueHandler, gdoer, kiwiServer]

    return doers


class AdminProofHandler(doing.Doer):
    def __init__(self, hab, controller, mbx, verifier, wiq, proofs=None, **kwa):
        self.hab = hab
        self.controller = controller
        self.mbx = mbx
        self.verifier = verifier
        self.presentations = proofs if proofs is not None else decking.Deck()
        self.wiq = wiq
        super(AdminProofHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle proofs presented externally

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """
        yield  # enter context

        while True:
            while self.presentations:
                (pre, presentation) = self.presentations.popleft()
                vc = presentation["vc"]
                vcproof = bytearray(presentation["proof"].encode("utf-8"))

                creder = proving.Credentialer(crd=vc)

                prefixer, seqner, diger, isigers = proving.parseProof(vcproof)
                status = self.verifier.verify(pre, creder, prefixer, seqner, diger, isigers)
                pl = dict(
                    pre=pre.qb64,
                    vc=vc,
                    status=status,
                )

                print("STORING VC PROOF FOR MY CONTROLLER", self.controller, pl)

                # TODO: Add SAID signature on exn, then sanction `fwd` envelope
                ser = exchanging.exchange(route="/cmd/presentation/proof", payload=pl)
                msg = bytearray(ser.raw)
                msg.extend(self.hab.sanction(ser))

                self.mbx.storeMsg(self.controller + "/credential", msg)

                yield

            yield


class AdminCueHandler(doing.DoDoer):
    """

    """

    def __init__(self, controller, hab, mbx, cues=None, **kwa):
        """

        Parameters:
            mbx is Mailboxer for saving messages for controller
            cues is cues Deck from external mailbox to process

        """
        self.controller = controller
        self.hab = hab
        self.mbx = mbx
        self.cues = cues if cues is not None else decking.Deck()

        super(AdminCueHandler, self).__init__(doers=[doing.doify(self.cueDo)], **kwa)

    def cueDo(self, tymth, tock=0.0, **opts):
        """

        Handle cues coming out of our external Mailbox listener and forward to controller
        mailbox if appropriate

        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.cues:
                cue = self.cues.popleft()
                cueKin = cue["kin"]  # type or kind of cue
                if cueKin in ("delegating",):
                    srdr = cue["serder"]
                    msg = self.hab.interact(data=[
                        dict(i=srdr.pre, s=srdr.ked["s"], d=srdr.dig)
                    ])
                    witRctDoer = agenting.WitnessReceiptor(hab=self.hab, msg=msg)
                    self.extend([witRctDoer])

                    while not witRctDoer.done:
                        yield self.tock

                    self.remove([witRctDoer])
                elif cueKin in ("psUnescrow",):
                    srdr = cue["serder"]
                    wits = srdr.ked["b"]
                    witq = agenting.WitnessInquisitor(hab=self.hab, klas=agenting.HttpWitnesser, wits=wits)
                    self.extend([witq])

                    while srdr.pre not in self.hab.kevers:
                        witq.query(pre=srdr.pre)
                        yield 1.0

                    print("Successfully deletated to", srdr.pre)


                yield self.tock
            yield self.tock




# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands.multisig module

"""

import argparse

from keri import kering
from keri.app import keeping, habbing
from keri.app.cli.common import displaying
from keri.db import basing, dbing

parser = argparse.ArgumentParser(description='list group identifiers')
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.set_defaults(handler=lambda args: listGroupIdentifiers(args))


def listGroupIdentifiers(args):
    """
    Lists any group identifiers that the identifier represented by the provided name
    belongs to.

    Parameters:
        args (parseargs):  command line parameters

    """

    name = args.name
    with basing.openDB(name=name, temp=False, reload=True) as db, \
            keeping.openKS(name=name, temp=False) as ks:
        try:

            hab = habbing.Habitat(name=name, ks=ks, db=db, temp=False, create=False)
            displaying.printGroups(hab)
        except kering.ConfigurationError:
            print(f"prefix for {name} does not exist, incept must be run first", )





# -*- encoding: utf-8 -*-
"""
keri.app.cli.commands module

"""
import argparse
import keri

parser = argparse.ArgumentParser(description='Print version of KLI')
parser.set_defaults(handler=lambda args: version(args))


def version(args):
    print(keri.__version__)



# -*- encoding: utf-8 -*-
"""
keri.kli.commands.watcher module

"""
import argparse

from hio.base import doing

from keri import help
from keri.app import directing, watching
from keri.app.cli.common import existing

parser = argparse.ArgumentParser(description='Rotate watcher prefix')
parser.set_defaults(handler=lambda args: rotateWatcher(args))
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument("--watcher", "-w", help="QB64 identifier prefix of watcher to rotate", default="", required=True)

logger = help.ogler.getLogger()


def rotateWatcher(args):
    name = args.name
    wat = args.watcher

    watr = WatcherRotate(name=name, wat=wat)
    directing.runController(doers=[watr], expire=0.0)


class WatcherRotate(doing.DoDoer):

    def __init__(self, name, wat, **kwa):
        self.watcher = wat
        self.hab, doers = existing.openHabitat(name=name)
        self.rotr = watching.WatcherClientRotateDoer(hab=self.hab)
        doers.extend([self.rotr])
        self.toRemove = list(doers)

        doers.extend([doing.doify(self.rotateDo)])

        super(WatcherRotate, self).__init__(doers=doers, **kwa)


    def rotateDo(self, tymth, tock=0.0, **opts):
        # enter context
        yield self.tock

        self.rotr.msgs.append(self.watcher)

        while not self.rotr.cues:
            yield self.tock

        habr = self.hab.db.habs.get(self.hab.name)
        print("New Watcher Set:")
        for wat in habr.watchers:
            print("\t{}".format(wat))

        self.remove(self.toRemove)



# -*- encoding: utf-8 -*-
"""
KERI
keri.vdr.verifying module

VC verifier support
"""
import datetime
import logging
from typing import Type

from hio.help import decking

from .. import help, kering
from ..core import parsing, coring, scheming
from ..help import helping
from ..vc import proving
from ..vdr import eventing
from ..vdr.eventing import VcStates
from ..vdr.viring import Registry

logger = help.ogler.getLogger()


class Verifier:
    """
    Verifier class accepts and validates TEL events.

    """
    TimeoutPSE = 3600  # seconds to timeout partially signed credential escrow
    TimeoutMRE = 3600  # seconds to timeout missing registry escrows
    TimeoutMRI = 3600  # seconds to timeout missing issuer escrows
    TimeoutBCE = 3600  # seconds to timeout missing issuer escrows

    CredentialExpiry = 3600

    def __init__(self, hab, reger=None, creds=None, cues=None, **kwa):
        """
        Initialize Verifier instance

        Parameters:
            hab is Habitat for this verifier's context
            name is user synonym for this verifier
            reger is Registry database instance
            tevers is dict of Tever instances keys by registry identifier
        """
        self.hab = hab
        self.reger = reger if reger is not None else Registry(name=hab.name, temp=True)
        self.creds = creds if creds is not None else decking.Deck()  # subclass of deque
        self.cues = cues if cues is not None else decking.Deck()  # subclass of deque

        self.inited = False
        self.tvy = None
        self.psr = None

        # save init kwy word arg parameters as ._inits in order to later finish
        # init setup elseqhere after databases are opened if not below
        self._inits = kwa

        if self.hab.inited:
            self.setup()

    def setup(self):
        self.tvy = eventing.Tevery(reger=self.reger, db=self.hab.db, regk=None, local=False)
        self.psr = parsing.Parser(framed=True, kvy=self.hab.kvy, tvy=self.tvy)

        self.inited = True

    @property
    def tevers(self):
        """
        Returns .db.tevers
        """
        return self.reger.tevers

    def processMessages(self, creds=None):
        """
        Process message dicts in msgs or if msgs is None in .msgs
        Parameters:
            creds (Deck): each entry is dict that matches call signature of
                .processCredential
        """
        if creds is None:
            creds = self.creds

        while creds:
            self.processCredential(**creds.pull())

    def processCredential(self, creder, prefixer, seqner, diger, sigers):
        """
        Verify the data of the credential against the schema, the SAID of the credential and
        the CESR Proof on the credential and if valid, store the credential

        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential

        """
        regk = creder.status
        vcid = creder.said
        schema = creder.schema
        prov = creder.crd["p"]

        if regk not in self.tevers:  # registry event not found yet
            self.escrowMRE(creder, prefixer, seqner, diger, sigers)
            self.cues.append(dict(kin="query", q=dict(r="logs", pre=creder.issuer, sn=seqner)))
            self.cues.append(dict(kin="telquery", q=dict(r="tels", ri=regk, i=vcid)))
            raise kering.MissingRegistryError("registry identifier {} not in Tevers".format(regk))

        state, lastSeen = self.tevers[regk].vcState(vcid)
        if state is None:  # credential issuance event not found yet
            self.escrowMRE(creder, prefixer, seqner, diger, sigers)
            self.cues.append(dict(kin="query", q=dict(r="logs", pre=creder.issuer, sn=seqner)))
            self.cues.append(dict(kin="telquery", q=dict(r="tels", ri=regk, i=vcid)))
            raise kering.MissingRegistryError("credential identifier {} not in Tevers".format(vcid))

        if state is VcStates.expired:
            self.escrowMRE(creder, prefixer, seqner, diger, sigers)
            self.cues.append(dict(kin="query", q=dict(r="logs", pre=creder.issuer, sn=seqner)))
            self.cues.append(dict(kin="telquery", q=dict(r="tels", ri=regk, i=vcid)))
            raise kering.MissingRegistryError("credential identifier {} is out of date".format(vcid))
        elif state is VcStates.revoked:  # no escrow, credential has been revoked
            raise kering.InvalidCredentialStateError("credential {} in registrying is not in issued state".format(vcid,
                                                                                                                  regk))

        # we don't know about this issuer
        if creder.issuer not in self.hab.kevers:
            self.escrowMIE(creder, prefixer, seqner, diger, sigers)
            self.cues.append(dict(kin="query", q=dict(r="logs", pre=creder.issuer, sn=seqner)))
            raise kering.MissingIssuerError("issuer identifier {} not in Kevers".format(creder.issuer))

        # Verify the credential against the schema
        scraw = scheming.jsonSchemaCache.resolve(schema)
        if not scraw:
            self.escrowMSE(creder, prefixer, seqner, diger, sigers)
            self.cues.append(dict(kin="query", q=dict(r="schema", said=schema)))
            raise kering.MissingSchemaError("schema {} not in cache".format(schema))


        schemer = scheming.Schemer(raw=scraw)
        if not schemer.verify(creder.raw):
            raise kering.FailedSchemaValidationError("Credential {} is not valid against schema {}"
                                                     .format(creder.said, schema))

        # Verify the signatures are valid and that the signature threshold as of the signing event is met
        tholder, verfers = self.hab.verifiage(pre=prefixer.qb64, sn=seqner.sn, dig=diger.qb64)
        _, indices = eventing.verifySigs(creder, sigers, verfers)

        if not tholder.satisfy(indices):  # We still don't have all the sigers, need to escrow
            self.escrowPSC(creder, prefixer, seqner, diger, sigers)
            raise kering.MissingSignatureError("Failure satisfying credential sith = {} on sigs for {}"
                                               " for evt = {}.".format(tholder.sith,
                                                                       [siger.qb64 for siger in sigers],
                                                                       creder.crd))
        for s in prov:
            for label, node in s.items():
                nodeSubject = node["i"]
                nodeSaid = node["d"]
                status = self.verifyChain(label, nodeSubject, nodeSaid)
                if status is None:
                    self.escrowMCE(creder, prefixer, seqner, diger, sigers)
                    self.cues.append(dict(kin="proof", subject=nodeSubject, said=nodeSaid))
                    raise kering.MissingChainError("Failure to verify credential {} chain {}({}) for {}"
                                                   .format(creder.said, label, nodeSaid, nodeSubject))
                elif status == VcStates.expired:
                    self.escrowMCE(creder, prefixer, seqner, diger, sigers)
                    self.cues.append(dict(kin="query", q=dict(r="tels", pre=nodeSaid)))
                    raise kering.MissingChainError("Failure to verify credential {} chain {}({}) for {}"
                                                   .format(creder.said, label, nodeSaid, nodeSubject))
                elif status == VcStates.revoked:
                    raise kering.RevokedChainError("Failure to verify credential {} chain {}({}) for {}"
                                                   .format(creder.said, label, nodeSaid, nodeSubject))
                else:  # VcStatus == VcStates.Issued
                    logger.info("Successfully validated credential chain {} for credential {}"
                                .format(label, creder.said))


        self.saveCredential(creder, prefixer, seqner, diger, sigers)
        proof = proving.buildProof(prefixer, seqner, diger, sigers)
        self.cues.append(dict(kin="saved", creder=creder, proof=proof))

    def escrowPSC(self, creder, prefixer, seqner, diger, sigers):
        """
        Credential Partial Signature Escrow


        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential

        """
        key = creder.saider.qb64b

        self._persist(creder, prefixer, seqner, diger, sigers)
        self.reger.pse.pin(keys=key, val=coring.Dater())

    def escrowMRE(self, creder, prefixer, seqner, diger, sigers):
        """
        Missing Registry Escrow


        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential

        """
        key = creder.saider.qb64b

        self._persist(creder, prefixer, seqner, diger, sigers)
        self.reger.mre.pin(keys=key, val=coring.Dater())

    def escrowMIE(self, creder, prefixer, seqner, diger, sigers):
        """
        Missing Issuer Escrow


        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential

        """
        key = creder.saider.qb64b

        self._persist(creder, prefixer, seqner, diger, sigers)
        self.reger.mie.pin(keys=key, val=coring.Dater())

    def escrowMCE(self, creder, prefixer, seqner, diger, sigers):
        """
        Missing Issuer Escrow


        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential

        """
        key = creder.saider.qb64b

        self._persist(creder, prefixer, seqner, diger, sigers)
        self.reger.mce.pin(keys=key, val=coring.Dater())

    def escrowMSE(self, creder, prefixer, seqner, diger, sigers):
        """
        Missing Credential Schema Escrow


        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential

        """
        key = creder.saider.qb64b

        self._persist(creder, prefixer, seqner, diger, sigers)
        self.reger.mse.pin(keys=key, val=coring.Dater())

    def processEscrows(self):
        """
        Process all escrows once each

        """

        self._processEscrow(self.reger.mce, self.TimeoutMRI, kering.MissingChainError)
        self._processEscrow(self.reger.mse, self.TimeoutMRI, kering.MissingSchemaError)
        self._processEscrow(self.reger.pse, self.TimeoutPSE, kering.MissingSignatureError)
        self._processEscrow(self.reger.mie, self.TimeoutMRI, kering.MissingIssuerError)
        self._processEscrow(self.reger.mre, self.TimeoutMRE, kering.MissingRegistryError)

    def _processEscrow(self, db, timeout, etype: Type[Exception]):
        """
        Generic credential escrow processing

        Parameters:
            db:
            timeout:
            etype:

        """
        for (said,), dater in db.getItemIter():
            creder, prefixer, seqner, diger, sigers = self._read(said)

            try:

                dtnow = helping.nowUTC()
                dte = helping.fromIso8601(dater.dts)
                if (dtnow - dte) > datetime.timedelta(seconds=timeout):
                    # escrow stale so raise ValidationError which unescrows below
                    logger.info("Verifier unescrow error: Stale event escrow "
                                " at said = %s\n", bytes(said))

                    raise kering.ValidationError("Stale event escrow "
                                                 "at said = {}.".format(bytes(said)))

                self.processCredential(creder, prefixer, seqner, diger, sigers)

            except etype as ex:
                print(ex)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Verifiery unescrow failed: %s\n", ex.args[0])
                else:
                    logger.error("Verifier unescrow failed: %s\n", ex.args[0])
            except Exception as ex:  # log diagnostics errors etc
                print(ex)
                # error other than missing sigs so remove from PA escrow
                db.rem(said)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Verifier unescrowed: %s\n", ex.args[0])
                else:
                    logger.error("Verifier unescrowed: %s\n", ex.args[0])
            else:
                db.rem(said)
                logger.info("Verifier unescrow succeeded in valid group op: "
                            "creder=\n%s\n", creder.pretty())

    def saveCredential(self, creder, prefixer, seqner, diger, sigers):
        """
        Write the credential and associated indicies to the database

        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential
        """
        self._persist(creder, prefixer, seqner, diger, sigers)

        schema = creder.schema.encode("utf-8")
        issuer = creder.issuer.encode("utf-8")
        subject = creder.subject["i"].encode("utf-8")

        # Look up indicies
        saider = creder.saider
        self.reger.issus.add(keys=issuer, val=saider)
        self.reger.subjs.add(keys=subject, val=saider)
        self.reger.schms.add(keys=schema, val=saider)

    def _persist(self, creder, prefixer, seqner, diger, sigers):
        """
        Save the base credential and seals (est evt+sigs quad) with no indices.

        Parameters:
            creder (Credentialer) that contains the credential to process
            prefixer (Prefixer) Identifier of the issuer of the credential
            seqner (Seqner) is the sequence number of the event used to sign the credential
            diger (Diger) is the digest of the event used to sign the credential
            sigers (list) are the cryptographic signatures on the credential
        """
        key = creder.saider.qb64b
        self.reger.creds.put(keys=key, val=creder)

        # Signer KEL Location and signatures
        vals = [(prefixer, seqner, diger, siger) for siger in sigers]
        self.reger.seals.put(keys=key, vals=vals)

    def _read(self, said):
        """
        Load base credential and seals(est evt+sigs quad) from database.

        Parameters:
            said(str or bytes): qb64 SAID of credential

        """

        creder = self.reger.creds.get(keys=said)

        # TODO:  de-dupe the seals here and extract the signatures
        seals = self.reger.seals.get(keys=said)
        prefixer = None
        seqner = None
        diger = None
        sigers = []
        for seal in seals:
            (prefixer, seqner, diger, siger) = seal
            sigers.append(siger)

        return creder, prefixer, seqner, diger, sigers

    def query(self, regk, vcid, res, dt=None, dta=None, dtb=None):
        """
        Returns query message for querying for a single element of type res
        """
        kever = self.hab.kever
        serder = eventing.query(regk=regk, vcid=vcid, res=res, dt=dt, dta=dta, dtb=dtb)

        sigers = self.hab.mgr.sign(ser=serder.raw, verfers=kever.verfers)
        msg = bytearray(serder.raw)  # make copy into new bytearray so can be deleted

        msg.extend(coring.Counter(coring.CtrDex.TransLastIdxSigGroups, count=1).qb64b)
        msg.extend(self.hab.pre.encode("utf-8"))

        counter = coring.Counter(code=coring.CtrDex.ControllerIdxSigs,
                                 count=len(sigers))
        msg.extend(counter.qb64b)
        for siger in sigers:
            msg.extend(siger.qb64b)

        return msg

    def verifyChain(self, label, nodeSubject, nodeSaid):
        """
        Verifies the node credential at the end of an edge

        Returns: None if the credential is not found
                 Expired if the state is issued and the date on the TEL event has expired
                 Issued if the credential has been saved and issued
                 Revoked if the credential has been revoked.

        Parameters:
            label (str): edge label
            nodeSubject(str): qb64 of node credential subject
            nodeSaid: (str) qb64 SAID of node credential

        """
        creder = self.reger.creds.get(keys=nodeSaid)
        if creder is None:
            return None

        iss = self.reger.subjs.get(keys=nodeSubject)
        if iss is None:
            return None

        if creder.status not in self.tevers:
            return None

        tever = self.tevers[creder.status]

        status, lastSeen = tever.vcState(nodeSaid)
        if status is None:
            return None

        dtnow = helping.nowUTC()
        dte = helping.fromIso8601(lastSeen.dts)
        if (dtnow - dte) > datetime.timedelta(seconds=self.CredentialExpiry):
            return VcStates.Expired

        return status



# -*- encoding: utf-8 -*-
"""
keri.core.scheming module

self-addressing and schema support
"""

import json

import cbor2 as cbor
import jsonschema
import msgpack

from . import coring
from .coring import MtrDex, Serials, Saider, Ids
from .. import help
from ..kering import ValidationError, DeserializationError

logger = help.ogler.getLogger()


class CacheResolver:
    """
    Sample jsonschema resolver for loading schema $ref references from a local hash.

    """

    def __init__(self, cache=None):
        """
        Create a jsonschema resolver that can be used for loading references to schema remotely.

        Parameters:
            cache (dict) is an optional pre-loaded cache of schema
        """
        self.cache = cache if cache is not None else dict()

    def add(self, key, schema):
        """
        Add schema to cache for resolution

        Parameters:
            key (str) URI to resolve to the schema
            schema (bytes) is bytes of the schema for the URI
        """
        self.cache[key] = schema

    def resolve(self, uri):
        if uri not in self.cache:
            return None

        ref = self.cache[uri]
        return ref

    def handler(self, uri):
        """
        Handler provided to jsonschema for cache resolution

        Parameters:
            uri (str) the URI to resolve
        """
        ref = self.resolve(uri)
        if not ref:
            return None

        schemr = Schemer(raw=ref)
        return schemr.sed

    def resolver(self, scer=b''):
        """
        Returns a jsonschema resolver for returning locally cached schema based on self-addressing
        identifier URIs.

        Parameters:
            scer (bytes) is the source document that is being processed for reference resolution

        """
        return jsonschema.RefResolver("", scer, handlers={"did": self.handler})


jsonSchemaCache = CacheResolver(cache={
    "E3n2Od38xMVDoM6Km-Awse_Cw9z0RtUJN-j0MQo642xw": b'{"$id":"E3n2Od38xMVDoM6Km-Awse_Cw9z0RtUJN-j0MQo642xw",'
                                                    b'"$schema":"http://json-schema.org/draft-07/schema#",'
                                                    b'"title":"Legal Entity Official Organizational Role vLEI '
                                                    b'Credential","description":"A vLEI Role Credential issued by a '
                                                    b'Qualified vLEI issuer to official representatives of a Legal '
                                                    b'Entity","properties":{"v":{"type":"string"},'
                                                    b'"d":{"type":"string"},"i":{"type":"string"},'
                                                    b'"s":{"description":"schema SAID","type":"string"},'
                                                    b'"a":{"description":"data block","properties":{"d":{'
                                                    b'"type":"string"},"i":{"type":"string"},'
                                                    b'"dt":{"description":"issuance date time","format":"date-time",'
                                                    b'"type":"string"},"ri":{"description":"credential status '
                                                    b'registry","type":"string"},"LEI":{"type":"string"},'
                                                    b'"personLegalName":{"type":"string"},"officialRole":{'
                                                    b'"type":"string"},"t":{"contains":{'
                                                    b'"const":"LegalEntityOfficialOrganizationalRolevLEICredential"},'
                                                    b'"type":"array"}},"additionalProperties":false,"required":["i",'
                                                    b'"dt","ri","LEI","personLegalName","officialRole","t"],'
                                                    b'"type":"object"},"p":{"contains":{"type":"object"},'
                                                    b'"description":"source block","items":{"properties":{'
                                                    b'"qualifiedvLEIIssuervLEICredential":{"description":"chain to '
                                                    b'issuer credential","properties":{"d":{"type":"string"},'
                                                    b'"i":{"type":"string"}},"additionalProperties":false,'
                                                    b'"type":"object"}},"additionalProperties":false,"required":['
                                                    b'"qualifiedvLEIIssuervLEICredential"],"type":"object"},'
                                                    b'"maxItems":1,"minItems":1,"type":"array"}},'
                                                    b'"additionalProperties":false,"required":["i","s","d"],'
                                                    b'"type":"object"}',
    "EJEY6JAAVfAh8-yBTV37rHaJ9b_VKvkZunz_oJupzsvQ": b'{"$id":"EJEY6JAAVfAh8-yBTV37rHaJ9b_VKvkZunz_oJupzsvQ",'
                                                    b'"$schema":"http://json-schema.org/draft-07/schema#",'
                                                    b'"title":"Legal Entity vLEI Credential","description":"A vLEI '
                                                    b'Credential issued by a Qualified vLEI issuer to a Legal '
                                                    b'Entity","properties":{"v":{"type":"string"},'
                                                    b'"d":{"type":"string"},"i":{"type":"string"},'
                                                    b'"s":{"description":"schema SAID","type":"string"},'
                                                    b'"a":{"description":"data block","properties":{"d":{'
                                                    b'"type":"string"},"i":{"type":"string"},'
                                                    b'"dt":{"description":"issuance date time","format":"date-time",'
                                                    b'"type":"string"},"ri":{"description":"credential status '
                                                    b'registry","type":"string"},"LEI":{"type":"string"},'
                                                    b'"t":{"contains":{"const":"LegalEntityvLEICredential"},'
                                                    b'"type":"array"}},"additionalProperties":false,"required":["i",'
                                                    b'"dt","ri","LEI","t"],"type":"object"},"p":{"contains":{'
                                                    b'"type":"object"},"description":"source block",'
                                                    b'"items":{"properties":{"qualifiedvLEIIssuervLEICredential":{'
                                                    b'"description":"chain to issuer credential","properties":{"d":{'
                                                    b'"type":"string"},"i":{"type":"string"}},'
                                                    b'"additionalProperties":false,"type":"object"}},'
                                                    b'"additionalProperties":false,"required":['
                                                    b'"qualifiedvLEIIssuervLEICredential"],"type":"object"},'
                                                    b'"maxItems":1,"minItems":1,"type":"array"}},'
                                                    b'"additionalProperties":false,"required":["i","s","d"],'
                                                    b'"type":"object"}',
    "EmaEqu_zIkxXKsrNJFTJq_s2c96McS8yzHhcvYDW8u5A": b'{"$id":"EmaEqu_zIkxXKsrNJFTJq_s2c96McS8yzHhcvYDW8u5A",'
                                                    b'"$schema":"http://json-schema.org/draft-07/schema#",'
                                                    b'"title":"Legal Entity Engagement Context Role vLEI Credential",'
                                                    b'"description":"A vLEI Role Credential issued to representatives '
                                                    b'of a Legal Entity in other than official roles but in '
                                                    b'functional or other context of engagement","properties":{"v":{'
                                                    b'"type":"string"},"d":{"type":"string"},"i":{"type":"string"},'
                                                    b'"s":{"description":"schema SAID","type":"string"},'
                                                    b'"a":{"description":"data block","properties":{"d":{'
                                                    b'"type":"string"},"i":{"type":"string"},'
                                                    b'"dt":{"description":"issuance date time","format":"date-time",'
                                                    b'"type":"string"},"ri":{"description":"credential status '
                                                    b'registry","type":"string"},"LEI":{"type":"string"},'
                                                    b'"personLegalName":{"type":"string"},"engagementContextRole":{'
                                                    b'"type":"string"},"t":{"contains":{'
                                                    b'"const":"LegalEntityEngagementContextRolevLEICredential"},'
                                                    b'"type":"array"}},"additionalProperties":false,"required":["i",'
                                                    b'"dt","ri","LEI","personLegalName","engagementContextRole","t"],'
                                                    b'"type":"object"},"p":{"contains":{"type":"object"},'
                                                    b'"description":"source block","items":{"properties":{'
                                                    b'"qualifiedvLEIIssuervLEICredential":{"description":"chain to '
                                                    b'issuer credential","properties":{"d":{"type":"string"},'
                                                    b'"i":{"type":"string"}},"additionalProperties":false,'
                                                    b'"type":"object"}},"additionalProperties":false,"required":['
                                                    b'"qualifiedvLEIIssuervLEICredential"],"type":"object"},'
                                                    b'"maxItems":1,"minItems":1,"type":"array"}},'
                                                    b'"additionalProperties":false,"required":["i","s","d"],'
                                                    b'"type":"object"}',
    # GLEIFvLEICredential
    "ES63gXI-FmM6yQ7ISVIH__hOEhyE6W6-Ev0cArldsxuc": b'{"$id":"ES63gXI-FmM6yQ7ISVIH__hOEhyE6W6-Ev0cArldsxuc",'
                                                    b'"$schema":"http://json-schema.org/draft-07/schema#",'
                                                    b'"title":"GLEIF vLEI Credential","description":"The vLEI '
                                                    b'Credential issued to GLEIF","type":"object","properties":{"v":{'
                                                    b'"type":"string"},"d":{"type":"string"},"i":{"type":"string"},'
                                                    b'"s":{"description":"schema SAID","type":"string"},'
                                                    b'"a":{"description":"data block","properties":{"d":{'
                                                    b'"type":"string"},"i":{"type":"string"},'
                                                    b'"dt":{"description":"issuance date time","format":"date-time",'
                                                    b'"type":"string"},"ri":{"description":"credential status '
                                                    b'registry","type":"string"},"LEI":{"type":"string"},'
                                                    b'"t":{"contains":{"const":"GLEIFvLEICredential"},'
                                                    b'"type":"array"}},"additionalProperties":false,"required":["d",'
                                                    b'"dt","ri","LEI","t"],"type":"object"},"p":{"maxItems":0,'
                                                    b'"minItems":0,"type":"array"}},"additionalProperties":false,'
                                                    b'"required":["d","i"]}',
    # QualifiedvLEIIssuervLEICredential
    "E-_XCbf1LJ0v9CR7g-_gOknf5dpoZROgF7qG5T8mXCv8": b'{"$id":"E-_XCbf1LJ0v9CR7g-_gOknf5dpoZROgF7qG5T8mXCv8",'
                                                    b'"$schema":"http://json-schema.org/draft-07/schema#",'
                                                    b'"title":"Qualified vLEI Issuer Credential","description":"A '
                                                    b'vLEI Credential issued by GLEIF to Qualified vLEI Issuers which '
                                                    b'allows the Qualified vLEI Issuers to issue, verify and revoke '
                                                    b'Legal Entity vLEI Credentials and Legal Entity Official '
                                                    b'Organizational Role vLEI Credentials","properties":{"v":{'
                                                    b'"type":"string"},"d":{"type":"string"},"i":{"type":"string"},'
                                                    b'"s":{"description":"schema SAID","type":"string"},'
                                                    b'"a":{"description":"data block","properties":{"d":{'
                                                    b'"type":"string"},"i":{"type":"string"},'
                                                    b'"dt":{"description":"issuance date time","format":"date-time",'
                                                    b'"type":"string"},"ri":{"description":"credential status '
                                                    b'registry","type":"string"},"LEI":{"type":"string"},'
                                                    b'"gracePeriod":{"default":90,"type":"integer"},"t":{"contains":{'
                                                    b'"const":"QualifiedvLEIIssuervLEICredential"},"type":"array"}},'
                                                    b'"additionalProperties":false,"required":["i","dt","ri","LEI",'
                                                    b'"t"],"type":"object"},"p":{"maxItems":0,"minItems":0,'
                                                    b'"type":"array"}},"additionalProperties":false,"required":["i",'
                                                    b'"d"],"type":"object"}',
})


class JSONSchema:
    """
    JSON Schema support class
    """
    id_ = Ids.dollar  # ID Field Label

    def __init__(self, resolver=CacheResolver()):
        self.resolver = resolver

    def resolve(self, uri):
        return self.resolver.resolve(uri)

    def load(self, raw=b'', kind=Serials.json):
        if kind == Serials.json:
            try:
                sed = json.loads(raw.decode("utf-8"))
            except Exception as ex:
                raise DeserializationError("Error deserializing JSON: {} {}"
                                           "".format(raw.decode("utf-8"), ex))

        elif kind == Serials.mgpk:
            try:
                sed = msgpack.loads(raw)
            except Exception as ex:
                raise DeserializationError("Error deserializing MGPK: {} {}"
                                           "".format(raw, ex))

        elif kind == Serials.cbor:
            try:
                sed = cbor.loads(raw)
            except Exception as ex:
                raise DeserializationError("Error deserializing CBOR: {} {}"
                                           "".format(raw, ex))
        else:
            raise ValueError("Invalid serialization kind = {}".format(kind))

        if self.id_ in sed:
            saider = Saider(qb64=sed[self.id_], label=self.id_)
            said = sed[self.id_]
            if not saider.verify(sed, prefixed=True, kind=kind, label=self.id_):
                raise ValidationError("invalid self-addressing identifier {} instead of {} in schema = {}"
                                      "".format(said, saider.qb64, sed))
        else:
            raise ValidationError("missing ID field {} in schema = {}"
                                  "".format(self.id_, sed))

        return sed, kind, saider

    @staticmethod
    def dump(sed, kind=Serials.json):
        raw = coring.dumps(sed, kind)
        return raw

    @staticmethod
    def detect(raw=b''):
        """
        Returns True if content represents JSON Schema by checking
            for $schema;  False otherwise
        """

        try:
            raw.index(b'"$schema"')
        except ValueError:
            return False

        return True

    @staticmethod
    def verify_schema(schema):
        """
        Returns True if the provided schema validates successfully
          as complaint Draft 7 JSON Schema False otherwise

        Parameters:
            schema (dict): is the JSON schema to verify
        """
        try:
            jsonschema.Draft7Validator.check_schema(schema=schema)
        except jsonschema.exceptions.SchemaError:
            return False

        return True

    def verify_json(self, schema=b'', raw=b''):
        """
        Returns True if the JSON passes validation against the
           provided complaint Draft 7 JSON Schema.  Returns False
           if raw is not valid JSON, schema is not valid JSON Schema or
           the validation fails

        Parameters:
              schema (bytes): is the schema use for validation
              raw (bytes): is JSON to validate against the Schema
        """
        try:
            d = json.loads(raw)
            jsonschema.validate(instance=d, schema=schema, resolver=self.resolver.resolver(scer=raw))
        except jsonschema.exceptions.ValidationError as ex:
            logger.error(f'jsonschema.exceptions.ValidationError {ex}')
            return False
        except jsonschema.exceptions.SchemaError as ex:
            logger.error(f'jsonschema.exceptions.SchemaError {ex}')
            return False
        except json.decoder.JSONDecodeError as ex:
            logger.error(f'json.decoder.JSONDecodeError {ex}')
            return False
        except Exception:
            return False

        return True


class Schemer:
    """
    Schemer is KERI schema serializer-deserializer class
    Verifies self-addressing identifier base on schema type
    Only supports current version VERSION

    Has the following public properties:

    Properties:
        .raw is bytes of serialized event only
        .sed is JSON schema dict
        .kind is Schema kind string value (see namedtuple coring.Serials)
        .saider is Saider instance of self-addressing identifier
        .said  is qb64 digest from .saider

    Hidden Attributes:
          ._raw is bytes of serialized schema only
          ._sed is JSON schema dict
          ._kind is schema kind string value (see namedtuple coring.Serials)
            supported kinds are 'JSONSchema'
          ._code is default code for .saider
          ._saider is Saider instance of digest of .raw


    """

    def __init__(self, raw=b'', sed=None, kind=None, typ=JSONSchema(), code=MtrDex.Blake3_256):
        """
        Deserialize if raw provided
        Serialize if sed provided but not raw
        When serilaizing if kind provided then use kind instead of field in sed

        Parameters:
          raw is bytes of serialized schema
          sed is JSON dict or None
            if None its deserialized from raw
          schemaType is the type of schema
          kind is serialization kind string value or None (see namedtuple coring.Serials)
            supported kinds are 'json', 'cbor', 'msgpack', 'binary'
            if kind is None then its extracted from ked or raw
          code is .saider default digest code

        """

        self._code = code
        if raw:
            self.raw = raw
        elif sed:
            self.typ = typ
            self._kind = kind
            self.sed = sed
        else:
            raise ValueError("Improper initialization need raw or sed.")

        if not self._verify_schema():
            raise ValidationError("invalid kind {} for schema {}"
                                  "".format(self.kind, self.sed))

    def _inhale(self, raw):
        """
        Loads type specific Schema ked and verifies the self-addressing identifier
            of the raw content

        Parameters:
            raw: JSON to load

        """
        self.typ = self._sniff(raw)
        sed, kind, saider = self.typ.load(raw=raw)

        return sed, kind, saider

    def _exhale(self, sed, kind=None):
        """
        Dumps type specific Schema JSON and returns the raw bytes, sed
           and schema kind

        Parameters:
            sed: JSON to load
            kind (Schema) tuple of schema type

        """
        saider = Saider(sad=sed, code=self._code, label=self.typ.id_)
        sed[self.typ.id_] = saider.qb64
        raw = self.typ.dump(sed)

        return raw, sed, kind, saider

    @staticmethod
    def _sniff(raw):
        try:
            raw.index(b'"$schema"')
        except ValueError:
            pass
        else:
            return JSONSchema()

        # Default for now is JSONSchema because we don't support any other
        return JSONSchema()

    @property
    def raw(self):
        """ raw property getter """
        return self._raw

    @raw.setter
    def raw(self, raw):
        """ raw property setter """
        sed, kind, saider = self._inhale(raw=raw)
        self._raw = bytes(raw)  # crypto ops require bytes not bytearray
        self._sed = sed
        self._kind = kind
        self._saider = saider

    @property
    def sed(self):
        """ ked property getter"""
        return self._sed

    @sed.setter
    def sed(self, sed):
        """ ked property setter  assumes ._kind """
        raw, sed, kind, saider = self._exhale(sed=sed, kind=self._kind)
        self._raw = raw
        self._kind = kind
        self._sed = sed
        self._saider = saider

    @property
    def kind(self):
        """ kind property getter """
        return self._kind

    @kind.setter
    def kind(self, kind):
        """ kind property setter Assumes ._ked """
        raw, kind, sed, saider = self._exhale(sed=self._sed, kind=kind)
        self._raw = raw
        self._sed = sed
        self._kind = kind
        self._saider = Saider(raw=self._raw, code=self._code, label=Ids.dollar)

    @property
    def saider(self):
        """ saider property getter """
        return self._saider

    @property
    def said(self):
        """ said property getter, relies on saider """
        return self.saider.qb64

    def verify(self, raw=b''):
        """
        Returns True if derivation from ked for .code matches .qb64 and
                If prefixed also verifies ked["i"] matches .qb64
                False otherwise

        Parameters:
            raw (bytes): is serialised JSON content to verify against schema
        """

        return self.typ.verify_json(schema=self.sed, raw=raw)

    def _verify_schema(self):
        """
        Returns True if derivation from ked for .code matches .qb64 and
                If prefixed also verifies ked["i"] matches .qb64
                False otherwise

        """

        return self.typ.verify_schema(schema=self.sed)



# -*- encoding: utf-8 -*-
"""
keri.vc.walleting module

"""
from hio.base import doing

from .. import help
from ..app import agenting
from ..core.scheming import CacheResolver

# TODO: create this and populate with needed schema for now
from ..vdr import viring

cache = CacheResolver()

logger = help.ogler.getLogger()


class Wallet:
    """
    Wallet represents all credentials received or verified


    """

    def __init__(self, db: viring.Registry = None, name="test", temp=False):
        """
        Create a Wallet associated with a Habitat

        Parameters:
            db: (viring.Registry) the database for the wallet

        """
        self.name = name
        self.temp = temp

        self.db = db if db is not None else viring.Registry(name=self.name, temp=self.temp)

    def getCredentials(self, schema=None):
        """
        Return list of (creder, prefixer, seqner, diger, sigers) for each credential
        that matches schema

        Parameters:
            schema: qb64 SAID of the schema for the credential

        """
        saiders = self.db.schms.get(keys=schema.encode("utf-8"))

        creds = []
        for saider in saiders:
            creder = self.db.creds.get(keys=saider.qb64b)

            # TODO:  de-dupe the seals here and extract the signatures
            seals = self.db.seals.get(keys=saider.qb64b)
            prefixer = None
            seqner = None
            diger = None
            sigers = []
            for seal in seals:
                (prefixer, seqner, diger, siger) = seal
                sigers.append(siger)

            print("found match")
            print(creder.pretty())

            creds.append((creder, prefixer, seqner, diger, sigers))

        return creds


class WalletDoer(doing.DoDoer):

    def __init__(self, hab, verifier, **kwa):
        """
        Wallter doer processes the verifier cues and escrows for an Enterprise Wallet

        Parameters:
            hab (Habitat) is the local environment associate with this wallet
            verifier (Verifier) is the verifier that processes and stores credentials

        """

        self.verifier = verifier

        doers = [doing.doify(self.escrowDo)]
        self.witq = agenting.WitnessInquisitor(hab=hab, klas=agenting.TCPWitnesser)


        super(WalletDoer, self).__init__(doers=doers, **kwa)


    def escrowDo(self, tymth, tock=0.0):
        """
        Returns:  doifiable Doist compatible generator method

        Usage:
            add result of doify on this method to doers list

        Processes the Groupy escrow for group icp, rot and ixn request messages.

        """
        # start enter context
        yield  # enter context
        while True:
            self.verifier.processEscrows()
            yield self.tock

    def verifierDo(self, tymth, tock=0.0, **opts):
        """
        Process cues from Verifier coroutine

            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters
        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.verifier.cues:
                cue = self.verifier.cues.popleft()
                cueKin = cue["kin"]

                if cueKin == "saved":
                    creder = cue["creder"]

                    logger.info("Credential: %s, Schema: %s,  Saved", creder.said, creder.schema)
                    logger.info(creder.pretty())

                elif cueKin == "query":
                    qargs = cue["q"]
                    self.witq.query(**qargs)

                elif cueKin == "telquery":
                    qargs = cue["q"]
                    self.witq.telquery(**qargs)
                yield self.tock
            yield self.tock



import argparse

from keri import kering
from keri.app import habbing

parser = argparse.ArgumentParser(description='List set of watchers for this environment (Habitat)')
parser.set_defaults(handler=lambda args: listWatchers(args))
parser.add_argument('--name', '-n', help='Human readable reference', required=True)


def listWatchers(args):
    name = args.name

    try:
        with habbing.existingHab(name=name) as hab:
            habr = hab.db.habs.get(name)
            print("Watcher Set:")
            for wat in habr.watchers:
                print("\t{}".format(wat))

    except kering.ConfigurationError:
        print(f"identifier prefix for {name} does not exist, incept must be run first", )
        return -1




import argparse

from hio import help
from hio.base import doing

from keri.app import directing, indirecting
from keri.app.cli.common import existing
from keri.vdr import registering

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Initialize a prefix')
parser.set_defaults(handler=lambda args: registryIncept(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--registry-name', '-r', help='Human readable name for registry, defaults to name of Habitat',
                    default=None)
parser.add_argument("--no-backers", "-nb", help="do not allow setting up backers different from the ahcnoring KEL "
                                                "witnesses", default=True, action="store")
parser.add_argument('--backers', '-b', help='New set of backers different from the anchoring KEL witnesses.  Can '
                                            'appear multiple times', metavar="<prefix>", default=[], action="append",
                    required=False)
parser.add_argument("--establishment-only", "-e", help="Only allow establishment events for the anchoring events of "
                                                       "this registry", default=False, action="store")


def registryIncept(args):
    name = args.name
    registryName = args.registry_name if args.registry_name is not None else name
    estOnly = args.establishment_only
    noBackers = args.no_backers
    backers = args.backers

    if noBackers and backers:
        print("--no-backers and --backers can not both be provided")
        return -1

    icpDoer = RegistryInceptor(name=name, registryName=registryName, estOnly=estOnly, noBackers=noBackers, baks=backers)

    doers = [icpDoer]
    directing.runController(doers=doers, expire=0.0)


class RegistryInceptor(doing.DoDoer):
    """

    """
    def __init__(self, name, registryName, **kwa):
        """


        """
        self.name = name
        self.registryName = registryName
        self.hab, doers = existing.openHabitat(name=self.name)

        mbx = indirecting.MailboxDirector(hab=self.hab, topics=["/receipt", "/multisig", "/replay"])
        self.icpr = registering.RegistryInceptDoer(hab=self.hab)
        doers.extend([self.icpr, mbx])
        self.toRemove = list(doers)

        doers.extend([doing.doify(self.inceptDo)])
        super(RegistryInceptor, self).__init__(doers=doers, **kwa)


    def inceptDo(self, tymth, tock=0.0, **kwa):
        """
        """
        yield self.tock

        msg = dict(name=self.registryName)
        self.icpr.msgs.append(msg)

        regk = None
        while not regk:
            while self.icpr.cues:
                cue = self.icpr.cues.popleft()
                if cue["kin"] == "finished":
                    regk = cue["regk"]
                    break
                yield self.tock
            yield self.tock

        print("Regsitry:  {}({}) \n\tcreated for Identifier Prefix:  {}".format(self.registryName, regk, self.hab.pre))

        self.remove(self.toRemove)



# -*- encoding: utf-8 -*-
"""
KERI
keri.app.grouping module

module for enveloping and forwarding KERI message
"""
import json
import logging
from collections import namedtuple

import blake3
import math
from hio import help
from hio.base import doing
from hio.help import decking

from keri import kering
from keri.app import agenting, forwarding
from keri.core import coring, eventing, parsing
from keri.db import dbing, basing
from keri.peer import exchanging
from keri.vc import proving

logger = help.ogler.getLogger()

Opage = namedtuple("Opage", 'icp rot ixn')

Ops = Opage(icp='icp', rot='rot', ixn='ixn')


class Groupy:
    """
    Processor of request messages for performing icp/rot/ixn events on multisig group identifiers

    """

    def __init__(self, hab, msgs=None, cues=None):
        """

        Parameters:
            hab(Habitat): environment of the local participant in multisig group
            msgs(decking.Deck): input messages of requests to perform operations on multisig group identifier
            cues(decking.Deck): output cues from processing messages

        """
        self.msgs = msgs if msgs is not None else decking.Deck()  # subclass of deque
        self.cues = cues if cues is not None else decking.Deck()  # subclass of deque
        self.hab = hab
        self.db = hab.db

        self.kvy = eventing.Kevery(db=hab.db,
                                   lax=False,
                                   local=False)

    def processMessages(self, msgs=None):
        """
        Process message dicts in msgs or if msgs is None in .msgs
        Parameters:
            msgs (Deck): each entry is dict that matches call signature of
                .processMessage
        """
        if msgs is None:
            msgs = self.msgs

        while msgs:
            self.processMessage(**msgs.pull())

    def processMessage(self, msg, mssrdr=None, sigers=None):
        """
        Process request message from controller to participant agent in form of dict
         for a icp/rot/ixn on the multisig group.

        Parameters:
            msg(dict): the request to incept/rotate/iteract this multisig group
            mssrdr(Serder): the multisig KEL event for this request
            sigers(list): list of Siger signatures on event

        """

        op = msg["op"]
        if op in (Ops.icp,):
            aids = list(msg['aids'])
            if self.hab.pre not in aids:
                raise kering.ConfigurationError("Local identifer {} must be member of aids ={}"
                                                .format(self.hab.pre, aids))

            sith = msg["isith"]
            toad = msg["toad"]
            nsith = msg["nsith"]
            wits = msg["witnesses"] if "witnesses" in msg is not None else self.hab.kever.wits
            data = msg["data"] if "data" in msg else None
            delpre = msg["delpre"] if "delpre" in msg else None

            if mssrdr is None:
                for aid in aids:
                    if aid not in self.hab.kevers:
                        self.cues.append(dict(kin="query", aid=aid))
                        self.escrowPAE(msg)
                        raise kering.MissingAidError(
                            "group {} missing AID {}".format(self.hab.pre, aid))

                mskeys, msdigers = self.extractKeysDigs(aids)

                if delpre:
                    mssrdr = eventing.delcept(keys=[mskey.qb64 for mskey in mskeys],
                                              sith=sith,
                                              toad=toad,
                                              wits=wits,
                                              nxt=coring.Nexter(digs=[diger.qb64 for diger in msdigers]).qb64,
                                              code=coring.MtrDex.Blake3_256,
                                              data=data,
                                              delpre=delpre)
                    sigers = self.signAndPropagate(mssrdr, aids)

                else:

                    mssrdr = eventing.incept(keys=[mskey.qb64 for mskey in mskeys],
                                             sith=sith,
                                             toad=toad,
                                             wits=wits,
                                             nxt=coring.Nexter(sith=nsith,
                                                               digs=[diger.qb64 for diger in msdigers]).qb64,
                                             code=coring.MtrDex.Blake3_256,
                                             data=data)

                    sigers = self.signAndPropagate(mssrdr, aids)

            indices = [siger.index for siger in sigers]
            tholder = coring.Tholder(sith=sith)
            if not tholder.satisfy(indices):  # We still don't have all the sigers, need to escrow
                self.escrowPSE(msg, mssrdr)
                raise kering.MissingSignatureError("Failure satisfying sith = {} on sigs for {}"
                                                   " for evt = {}.".format(tholder.sith,
                                                                           [siger.qb64 for siger in sigers],
                                                                           mssrdr.ked))

            #  Add this group identifier prefix to my list of group identifiers I participate in
            group = basing.GroupIdRecord(lid=self.hab.pre, gid=mssrdr.pre, dig=mssrdr.dig, cst=nsith, aids=aids)
            self.db.gids.pin(self.hab.pre, group)

            self.cues.append(dict(
                kin="logEvent",
                group=group,
                pre=mssrdr.pre,
                mssrdr=mssrdr,
                sigers=sigers,
                sn=0
            ))

        elif op in (Ops.rot,):
            sith = msg["sith"]
            toad = msg["toad"]
            data = msg["data"]
            reason = msg["reason"] if "reason" in msg else ""
            wits = msg["witnesses"] if "witnesses" in msg else []
            cuts = msg["witness_cut"] if "witnesse_cut" in msg else []
            adds = msg["witness_add"] if "witnesse_add" in msg else []

            group = self.hab.group()
            if group is None:
                print("invalid group identifier {}\n".format(self.hab.pre))
                raise kering.InvalidGroupError("invalid group identifier {}".format(self.hab.pre))

            others = list(group.aids)
            others.remove(group.lid)

            if sith is None:
                sith = "{:x}".format(max(0, math.ceil(len(group.aids) / 2)))

            if wits:
                if adds or cuts:
                    raise kering.ConfigurationError("you can only specify witnesses or cuts and adds")
                ewits = self.hab.kever.lastEst.wits

                cuts = set(wits) & set(ewits)
                adds = set(wits) - set(ewits)

            gkev = self.hab.kevers[group.gid]
            sno = gkev.sn + 1

            if self.hab.kever.sn == gkev.sn:  # We are equal to the current group identifier, need to rotate
                rot = self.hab.rotate()
                self.cues.append(dict(kin="witness", msg=bytearray(rot)))

                exn = exchanging.exchange(route="/multisig/event", payload=dict(evt=rot.decode("utf-8"), reason=reason))
                emsg = bytearray(exn.raw)
                emsg.extend(self.hab.sanction(serder=exn))
                self.cues.append(dict(kin="send", recipients=others, topic='multisig', evt=emsg,
                                      reason=reason))
                print("Local identifier rotated, checking other group members:")

            for aid in group.aids:
                kever = self.hab.kevers[aid]
                if aid != self.hab.pre:
                    if kever.sn < self.hab.kever.sn:
                        self.cues.append(dict(kin="query", aid=aid))
                        self.escrowPAE(msg)
                        raise kering.MissingAidError(
                            "Group {} - AID {} not at sn={}".format(self.hab.pre, aid, self.hab.kever.sn))

            if mssrdr is None:
                mskeys, msdigers = self.extractKeysDigs(group.aids)
                wits = gkev.wits
                mssrdr = eventing.rotate(pre=gkev.prefixer.qb64,
                                         dig=gkev.serder.dig,
                                         sn=sno,
                                         keys=[mskey.qb64 for mskey in mskeys],
                                         sith=group.cst,  # the previously committed to signing threshold
                                         toad=toad,
                                         wits=wits,
                                         cuts=cuts,
                                         adds=adds,
                                         data=data,
                                         nxt=coring.Nexter(sith=sith,  # the next digest previous calculated
                                                           digs=[diger.qb64 for diger in msdigers]).qb64)

                sigers = self.signAndPropagate(mssrdr, group.aids)

            indices = [siger.index for siger in sigers]
            if not mssrdr.tholder.satisfy(indices):  # If we still don't have all the sigers, need to escrow
                self.escrowPSE(msg, mssrdr)
                raise kering.MissingSignatureError("Failure satisfying sith = {} on sigs for {}"
                                                   " for evt = {}.".format(mssrdr.tholder.sith,
                                                                           [siger.qb64 for siger in sigers],
                                                                           mssrdr.ked))

            group.cst = sith
            group.dig = mssrdr.dig
            self.hab.db.gids.pin(self.hab.pre, group)

            self.cues.append(dict(
                kin="logEvent",
                group=group,
                pre=mssrdr.pre,
                mssrdr=mssrdr,
                sigers=sigers,
                sn=mssrdr.ked["s"]
            ))

        elif op in (Ops.ixn,):
            data = msg["data"]
            reason = msg["reason"] if "reason" in msg else ""

            group = self.hab.group()
            if group is None:
                print("invalid group identifier {}\n".format(self.hab.pre))
                raise kering.InvalidGroupError("invalid group identifier {}".format(self.hab.pre))

            others = list(group.aids)
            others.remove(group.lid)

            gkev = self.hab.kevers[group.gid]
            sno = gkev.sn + 1
            if self.hab.kever.sn == gkev.sn:  # We are equal to the current group identifier, need to interact
                ixn = self.hab.interact()
                exn = exchanging.exchange(route="/multisig/event", payload=dict(evt=ixn.decode("utf-8"), reason=reason))
                emsg = bytearray(exn.raw)
                emsg.extend(self.hab.sanction(serder=exn))

                self.cues.append(dict(kin="send", recipients=others, topic='multisig', evt=emsg,
                                      reason=reason))

                self.cues.append(dict(kin="witness", msg=bytearray(ixn)))

            if mssrdr is None:
                mssrdr = eventing.interact(pre=gkev.prefixer.qb64,
                                           dig=gkev.serder.dig,
                                           sn=sno,
                                           data=data)
                sigers = self.signAndPropagate(mssrdr, group.aids)

            indices = [siger.index for siger in sigers]
            if not gkev.tholder.satisfy(indices):  # If we still don't have all the sigers, need to escrow
                self.escrowPSE(msg, mssrdr)
                raise kering.MissingSignatureError("Failure satisfying sith = {} on sigs for {}"
                                                   " for evt = {}.".format(gkev.tholder.sith,
                                                                           [siger.qb64 for siger in sigers],
                                                                           mssrdr.ked))
            self.cues.append(dict(
                kin="logEvent",
                group=group,
                pre=mssrdr.pre,
                mssrdr=mssrdr,
                sigers=sigers,
                sn=mssrdr.ked["s"]
            ))

    def signAndPropagate(self, mssrdr, aids):
        """
        Sign message and cue up message to send to participants of the group as identified by
        the list in aids

        Parameters:
            mssrdr(Serder): is event to sign and send to participants in aids
            aids(list): list of qb64 identifier prefix of group participants

        """
        idx = aids.index(self.hab.pre)
        sigers = self.hab.mgr.sign(ser=mssrdr.raw, verfers=self.hab.kever.verfers, indices=[idx])
        msg = eventing.messagize(mssrdr, sigers=sigers)
        parsing.Parser().parseOne(ims=bytearray(msg), kvy=self.kvy)

        others = list(aids)
        others.remove(self.hab.pre)
        self.cues.append(dict(kin="send", recipients=others, topic='multisig', evt=bytearray(msg)))

        return sigers

    def escrowPAE(self, msg):
        """
        Partial AIDs Escrow

        Parameters:
            msg(dict): the request to incept/rotate/iteract this multisig group

        """
        dat = json.dumps(msg).encode("utf-8")
        self.db.gpae.add(self.hab.pre, dat)

    def escrowPSE(self, msg, mssrdr):
        """
        Partial Signature Escrow

        Parameters:
            msg(dict): the request to incept/rotate/iteract this multisig group
            mssrdr(Serder): the multisig KEL event for this request

        """
        msg["pre"] = mssrdr.pre
        msg["dig"] = mssrdr.dig
        dat = json.dumps(msg).encode("utf-8")
        self.db.gpse.add(self.hab.pre, dat)

    def processEscrows(self):
        """
        Process all escrows, once per call

        """
        self.processPartialAidEscrow()
        self.processPartialSignedEscrow()

    def processPartialAidEscrow(self):
        """
        Process escrow of multisig requests with missing participant AIDs.
        Message processing will query for the AID before commiting to escrow
        assuming that the AID will be in the KEL by the time we process this escrow.

        """
        for (pre,), mraw in self.db.gpae.getItemIter():
            try:
                msg = json.loads(mraw)
                self.processMessage(msg=msg)
            except kering.MissingAidError as ex:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Groupy unescrow failed: %s\n", ex.args[0])
                else:
                    logger.error("Groupy unescrow failed: %s\n", ex.args[0])
            except Exception as ex:  # log diagnostics errors etc
                # error other than missing AID so remove from PA escrow
                self.db.gpae.rem(pre, val=mraw)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Groupy unescrowed: %s\n", ex.args[0])
                else:
                    logger.error("Groupy unescrowed: %s\n", ex.args[0])
            else:
                self.db.gpae.rem(pre, val=mraw)
                logger.info("Groupy unescrow succeeded in valid group op: "
                            "msg=\n%s\n", json.dumps(msg, indent=1))

    def processPartialSignedEscrow(self):
        """
        Process escrow of partially signed multisig group KEL events.  Message
        processing will send this local controllers signature to all other participants
        so this escrow waits for signatures from all other participants

        """
        for (pre,), dat in self.db.gpse.getItemIter():
            msg = json.loads(dat)

            gid = msg["pre"]
            dig = msg["dig"]

            dgkey = dbing.dgKey(gid, dig)
            eraw = self.db.getEvt(dgkey)
            mssrdr = coring.Serder(raw=bytes(eraw))  # escrowed event

            dgkey = dbing.dgKey(mssrdr.preb, mssrdr.digb)
            sigs = self.hab.db.getSigs(dgkey)
            sigers = [coring.Siger(qb64b=bytes(sig)) for sig in sigs]

            try:
                self.processMessage(msg, mssrdr=mssrdr, sigers=sigers)
            except kering.MissingSignatureError as ex:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Groupy unescrow failed: %s\n", ex.args[0])
                else:
                    logger.error("Groupy unescrow failed: %s\n", ex.args[0])
            except Exception as ex:  # log diagnostics errors etc
                # error other than missing sigs so remove from PA escrow
                self.db.gpse.rem(pre)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Groupy unescrowed: %s\n", ex.args[0])
                else:
                    logger.error("Groupy unescrowed: %s\n", ex.args[0])
            else:
                self.db.gpse.rem(pre)
                logger.info("Groupy unescrow succeeded in valid group op: "
                            "msg=\n%s\n", json.dumps(msg, indent=1))

    def extractKeysDigs(self, aids):
        """
        Extract the public key and next digest from the current est event of the other
        participants in the multisig group.

        Parameters:
            aids(list): qb64 identifier prefix of all participants of the multisig group

        """
        mskeys = []
        msdigers = []
        for aid in aids:
            kever = self.hab.kevers[aid]
            keys = kever.verfers
            if len(keys) > 1:
                raise kering.ConfigurationError("Identifier must have only one key, {} has {}"
                                                .format(aid, len(keys)))

            diger = self.extractDig(nexter=kever.nexter, tholder=kever.tholder)

            mskeys.append(keys[0])
            msdigers.append(diger)

        return mskeys, msdigers

    @staticmethod
    def extractDig(nexter, tholder):
        """
        Extracts the original digest of the public key from the digest created by XORing the
        key with the signing threshold.  This is used in group identifier event creation to enable
        creation of the next digest with the combined keys and the group signing threshold.

        Parameters:
            nexter is Nexter instance of next sith and next signing keys
            tholder is Tholder instance for event sith

        """
        dint = int.from_bytes(nexter.raw, 'big')

        limen = tholder.limen
        ldig = blake3.blake3(limen.encode("utf-8")).digest()
        sint = int.from_bytes(ldig, 'big')
        kint = dint ^ sint

        diger = coring.Diger(raw=kint.to_bytes(coring.Matter._rawSize(coring.MtrDex.Blake3_256), 'big'))
        return diger


class MultiSigGroupDoer(doing.DoDoer):
    """
    DoDoer instance that launches the environment and dependencies needed to create and disseminate
    the inception event for a multisig group identifier.  The identifier of the environment loaded from `name`
    must be a member of the group of identifiers listed in the configuration file.

    """

    def __init__(self, hab, ims=None, msgs=None, cues=None):
        """
        Creates the DoDoer needed to modify a multisig group identifier.  Requires the
        name of the environment whose identifier is a member of the group being created.
        All other arguments are passed to the inceptDo generator method as parameters to create
        the inception event.

        Parameters
            name (str): Name of the local identifier environment

        """
        self.hab = hab
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()
        self.groupy = Groupy(hab=self.hab)
        self.ims = ims if ims is not None else bytearray()
        self.msgToSend = None

        self.postman = forwarding.Postman(hab=self.hab)
        self.witq = agenting.WitnessInquisitor(hab=self.hab, klas=agenting.TCPWitnesser)

        doers = [self.postman,
                 self.witq,
                 doing.doify(self.msgDo),
                 doing.doify(self.cueDo),
                 doing.doify(self.escrowDo),
                 ]

        super(MultiSigGroupDoer, self).__init__(doers=doers)

    def msgDo(self, tymth, tock=0.0):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list

        Message Parameters:
            group: str group name
            aids: list of qb64 identifier prefixes of participants in the group
            witnesses: list of qb64 identifier prefixes of witnesses for the group
            toad: str of witness receipt threshold for group KEL
            isith str of current signing threshold for group
            nsith str of next signing threshold for group
            data is list of dicts or anchors for the inception event of the group KEL

        """
        # start enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)  # finish enter context

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                try:
                    self.groupy.processMessage(msg=msg)
                except (kering.ValidationError, Exception) as ex:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.exception("Multisig message error: %s\n", ex)
                    else:
                        logger.error("Multisig message error: %s\n", ex)
                yield self.tock

            yield self.tock

    def cueDo(self, tymth, tock=0.0):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list

        Processes the cues coming out of the Groupy.  Performs queries, message
        delivery and receipting based on the cues from the Groupy processing of
        group requests.

        """
        # start enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)  # finish enter context

        while True:
            while self.groupy.cues:  # iteratively process each cue in cues
                cue = self.groupy.cues.popleft()
                cueKin = cue["kin"]  # type or kind of cue

                if cueKin == "query":
                    self.witq.query(cue["aid"])
                elif cueKin == "send":
                    recpts = cue["recipients"]
                    for recpt in recpts:
                        self.postman.send(recipient=recpt,
                                          topic=cue["topic"],
                                          msg=bytearray(cue["evt"]))
                elif cueKin == "witness":
                    msg = cue["msg"]
                    witRctDoer = agenting.WitnessReceiptor(hab=self.hab, msg=msg, klas=agenting.HttpWitnesser)
                    self.extend([witRctDoer])

                    while not witRctDoer.done:
                        _ = yield self.tock
                    self.remove([witRctDoer])

                elif cueKin == "logEvent":
                    group = cue["group"]
                    mssrdr = cue["mssrdr"]
                    sigers = cue["sigers"]
                    idx = group.aids.index(self.hab.pre)

                    if idx == sigers[0].index:  # We are the first signer, elected to send to witnesses

                        if mssrdr.ked["t"] in (coring.Ilks.dip, coring.Ilks.drt):
                            # Process event in local hab, send to delegator, don't receipt.
                            self.msgToSend = eventing.messagize(mssrdr, sigers=sigers)
                            self.ims.extend(bytearray(self.msgToSend))

                            di = mssrdr.ked["di"]
                            self.postman.send(recipient=di,
                                              topic="delegate",
                                              msg=bytearray(self.msgToSend))
                        else:

                            msg = eventing.messagize(mssrdr, sigers=sigers)
                            witRctDoer = agenting.WitnessReceiptor(hab=self.hab, msg=msg, klas=agenting.TCPWitnesser)
                            self.extend([witRctDoer])

                            while not witRctDoer.done:
                                _ = yield self.tock

                            self.remove([witRctDoer])
                            self.cues.append(cue)

                    else:  # We are not the first signer, so we wait for the sigs and processed receipts
                        while mssrdr.pre not in self.hab.kevers or self.hab.kevers[mssrdr.pre].sn < mssrdr.sn:
                            self.witq.query(mssrdr.pre)
                            _ = (yield self.tock)
                        self.cues.append(cue)

                yield self.tock

            yield self.tock

    def escrowDo(self, tymth, tock=0.0):
        """
        Returns:  doifiable Doist compatible generator method

        Usage:
            add result of doify on this method to doers list

        Processes the Groupy escrow for group icp, rot and ixn request messages.

        """
        # start enter context
        yield  # enter context
        while True:
            self.groupy.processEscrows()
            yield


class MultisigInceptHandler(doing.DoDoer):
    """
    Handler for multisig group inception notification EXN messages

    """
    resource = "/multisig/incept"


    def __init__(self, mbx, controller, cues=None, **kwa):
        """

        Parameters:
            mbx (Mailboxer) of format str names accepted for offers
            controller (str) qb64 identity prefix of controller
            cues (decking.Deck) of outbound cue messages from handler

        """
        self.controller = controller
        self.mbx = mbx
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        super(MultisigInceptHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a multisig/incept message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                pl = msg["payload"]
                pl["r"] = "/incept"
                raw = json.dumps(pl).encode("utf-8")
                self.mbx.storeMsg(self.controller+"/multisig", raw)

                yield
            yield


class MultisigIssueHandler(doing.DoDoer):
    """
    Handler for multisig group issuance notification EXN messages

    """
    resource = "/multisig/issue"


    def __init__(self, mbx, controller, cues=None, **kwa):
        """

        Parameters:
            mbx (Mailboxer) of format str names accepted for offers
            controller (str) qb64 identity prefix of controller
            cues (decking.Deck) of outbound cue messages from handler

        """
        self.controller = controller
        self.mbx = mbx
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        super(MultisigIssueHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a multisig/incept message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                pl = msg["payload"]
                pl["r"] = "/issue"
                raw = json.dumps(pl).encode("utf-8")
                self.mbx.storeMsg(self.controller+"/multisig", raw)

                yield
            yield


class MultisigEventHandler(doing.Doer):
    """
    Handler for multisig group rotation/interact notification EXN messages

    """

    resource = "/multisig/event"

    def __init__(self, hab, verifier, cues=None, **kwa):
        """

        Parameters:
            hab (Habitat) is environment of participant in multisig group
            controller (str) qb64 identity prefix of controller
            mbx (Mailboxer) of format str names accepted for offers
            cues (decking.Deck) of outbound cue messages from handler

        """
        self.hab = hab
        self.verifier = verifier
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        self.kvy = eventing.Kevery(db=self.hab.db, lax=False, local=False)

        super(MultisigEventHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a /multisig/interact message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        Payload:
            evt is bytes of ixn message from another participant
            reason is either a str expressing reason for interaction event or credential

        """
        yield self.tock

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                payload = msg["payload"]
                evt = payload["evt"].encode("utf-8")
                reason = payload["reason"]


                parsing.Parser().parse(bytearray(evt), kvy=self.kvy)
                if reason is not None:
                    craw = reason.encode("utf-8")
                    proving.parseCredential(ims=craw, verifier=self.verifier)

                yield

            yield



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse

from keri import kering
from keri.app import habbing
from keri.core import coring

parser = argparse.ArgumentParser(description='Sign an arbitrary string')
parser.set_defaults(handler=lambda args: handler(args))
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument("--prefix", "-p", help="Identifier prefix of the signer", required=True)
parser.add_argument('--text', '-t', help='Original signed text or file (starts with "@")', required=True)
parser.add_argument('--signature', '-s', default=[], help='list of signatures to verify (can appear multiple times)',
                    action="append", required=True)


def handler(args):

    name = args.name
    sigers = [coring.Siger(qb64=sig) for sig in args.signature]

    try:
        with habbing.existingHab(name=name) as hab:

            kever = hab.kevers[args.prefix]

            txt = args.text
            if txt.startswith("@"):
                f = open(txt[1:], "r")
                data = f.read()
            else:
                data = txt

            ser = data.encode("utf-8")
            verfers = kever.verfers
            for siger in sigers:
                if siger.index >= len(verfers):
                    raise kering.ValidationError("Index = {} to large for keys."
                                                 "".format(siger.index))
                siger.verfer = verfers[siger.index]  # assign verfer
                if siger.verfer.verify(siger.raw, ser):  # verify each sig
                    print("Signature {} is valid.".format(siger.index+1))
                else:
                    print("Signature {} is invalid.".format(siger.index+1))



    except kering.ConfigurationError:
        print(f"prefix for {name} does not exist, incept must be run first", )
    except FileNotFoundError:
        print("unable to open file", args.text[1:])



# -*- encoding: utf-8 -*-
"""
keri.kli.common.displaying module

"""

from keri.db import dbing


def printGroups(hab):
    """
    Print identifier and participant information for an environment (Habitat)
    Uses the .gids database to determine groups

    Parameters:
        hab (Habitat): environment for whom to print groups

    """
    group = hab.group()
    if group:
        printGroup(hab, group)


def printGroup(hab, group):
    """
    Print identifier and participant information for a specific group

    Parameters:
        hab (Habitat): environment for whom to print groups
        name (str): human readable name for group identifier
        group (GroupIdentifier): the local identifier and list of participants

    """
    print("Group: ", group.gid)
    printIdentifier(hab, group.gid)

    aids = group.aids
    parts = "Participants: \n" + ("\t{}\n" * len(aids))
    print(parts.format(*aids))
    print()


def printIdentifier(hab, pre):
    """
    Print current state information for the identifier prefix pre

    Parameters:
        hab (Habitat): environment that contains the information for the idenfitier prefix
        pre (str): qb64 of the identifier prefix
    :return:
    """
    kever = hab.kevers[pre]
    ser = kever.serder
    dgkey = dbing.dgKey(ser.preb, ser.digb)
    wigs = hab.db.getWigs(dgkey)

    print("Prefix:\t{}".format(pre))
    print("Seq No:\t{}".format(kever.sn))
    if kever.delegated:
        print("Delegated Identifier")

    print("\nWitnesses:")
    print("Count:\t\t{}".format(len(kever.wits)))
    print("Receipts:\t{}".format(len(wigs)))
    print("Threshold:\t{}".format(kever.toad))
    print("\nPublic Keys:\t")
    for idx, verfer in enumerate(kever.verfers):
        print(f'\t{idx+1}. {verfer.qb64}')
    print()



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands.delegate module

"""
import argparse
import json
import sys
from json import JSONDecodeError

from hio import help
from hio.base import doing

from keri import kering
from keri.app import directing, keeping, habbing, agenting
from keri.app.cli.commands.delegate.incept import DelegateOptions
from keri.core import eventing, coring, parsing
from keri.db import basing

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Initialize a seal for creating a delegated identifier')
parser.set_defaults(handler=lambda args: rotate(args))
parser.add_argument('--name', '-n', help='Human readable environment reference', required=True)
parser.add_argument('--file', '-f', help='Filename to use to create the identifier', default="", required=True)
parser.add_argument('--seal', '-s', help='Filename to write seal', default="", required=True)
parser.add_argument('--data', '-d', help='Anchor data, \'@\' allowed', default=None, action="store", required=True)


def rotate(args):
    """
    Reads the config file into a CreateDelegateOptions dataclass and creates
    delegated identifier prefixes and events

    Parameters:
        args: Parsed arguments from the command line

    """
    try:
        f = open(args.file)
        config = json.load(f)
        opts = DelegateOptions(**config)

    except FileNotFoundError:
        print("config file", args.file, "not found")
        sys.exit(-1)
    except JSONDecodeError:
        print("config file", args.file, "not valid JSON")
        sys.exit(-1)

    name = args.name
    sealFile = args.seal

    if args.data is not None:
        try:
            if args.data.startswith("@"):
                f = open(args.data[1:], "r")
                data = json.load(f)
            else:
                data = json.loads(args.data)
        except json.JSONDecodeError:
            raise kering.ConfigurationError("data supplied must be value JSON to anchor in a seal")

        if not isinstance(data, list):
            data = [data]

    else:
        data = None

    kwa = opts.__dict__
    icpDoer = DelegateRotateDoer(name=name, sealFile=sealFile, data=data, **kwa)

    doers = [icpDoer]
    directing.runController(doers=doers, expire=0.0)


class DelegateRotateDoer(doing.DoDoer):
    """
    DoDoer instance that launches the environment and dependencies needed to create and disseminate
    the inception event for a delegated identifier.
    """

    def __init__(self, name, sealFile, data, **kwa):
        """
        Creates the DoDoer needed to create the seal for a delegated identifier.

        Parameters
            name (str): Name of the local identifier environment

        """

        ks = keeping.Keeper(name=name, temp=False)  # not opened by default, doer opens
        self.ksDoer = keeping.KeeperDoer(keeper=ks)  # doer do reopens if not opened and closes
        db = basing.Baser(name=name, temp=False, reload=True)  # not opened by default, doer opens
        self.dbDoer = basing.BaserDoer(baser=db)  # doer do reopens if not opened and closes

        hab = habbing.Habitat(name=name, ks=ks, db=db, temp=False, create=False)
        self.habDoer = habbing.HabitatDoer(habitat=hab)

        doers = [self.ksDoer, self.dbDoer, self.habDoer, doing.doify(self.rotateDo, **kwa)]
        self.hab = hab
        self.data = data
        self.sealFile = sealFile
        self.delegatorPrefix = kwa["delegatorPrefix"]

        super(DelegateRotateDoer, self).__init__(doers=doers)

    def rotateDo(self, tymth, tock=0.0, **kwa):
        # start enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)  # finish enter context

        delPre = self.data[0]["i"]
        delK = self.hab.kevers[delPre]

        verfers, digers, cst, nst = self.hab.interact(pre=delK.prefixer.qb64, temp=False)
        rotSrdr = eventing.deltate(pre=delK.prefixer.qb64,
                                   keys=[verfer.qb64 for verfer in verfers],
                                   dig=delK.serder.diger.qb64,
                                   sn=delK.sn + 1,
                                   nxt=coring.Nexter(digs=[diger.qb64 for diger in digers]).qb64)

        seal = dict(i=rotSrdr.pre,
                    s=rotSrdr.ked["s"],
                    d=rotSrdr.dig)

        with open(self.sealFile, "w") as f:
            f.write(json.dumps(seal, indent=4))

        witq = agenting.WitnessInquisitor(hab=self.hab, klas=agenting.TCPWitnesser)
        self.extend([witq])

        print("Hello, could someone approve my delegated rotation, please?")

        while self.delegatorPrefix not in self.hab.kevers or self.hab.kevers[self.delegatorPrefix].sn < 2:
            witq.query(self.delegatorPrefix)
            yield self.tock

        sigers = self.hab.mgr.sign(ser=rotSrdr.raw, verfers=verfers)
        msg = bytearray(rotSrdr.raw)
        counter = coring.Counter(code=coring.CtrDex.ControllerIdxSigs,
                                 count=len(sigers))
        msg.extend(counter.qb64b)
        for siger in sigers:
            msg.extend(siger.qb64b)
        counter = coring.Counter(code=coring.CtrDex.SealSourceCouples,
                                 count=1)
        msg.extend(counter.qb64b)

        event = self.hab.kevers[rotSrdr.pre]
        seqner = coring.Seqner(sn=event.sn)
        msg.extend(seqner.qb64b)
        msg.extend(event.serder.diger.qb64b)

        delKvy = eventing.Kevery(db=self.hab.db, lax=True)
        parsing.Parser().parseOne(ims=bytearray(msg), kvy=delKvy)

        while rotSrdr.pre not in delKvy.kevers:
            yield self.tock

        print("Successfully rotated delegate identifier keys", rotSrdr.pre)
        print("Public key", rotSrdr.verfers[0].qb64)

        self.remove([self.ksDoer, self.dbDoer, self.habDoer, witq])

        return



# -*- encoding: utf-8 -*-
"""
keri.kli.commands.watcher module

"""
import argparse
import logging
import sys

import falcon
from hio.core import http
from hio.core.tcp import serving
from keri import help, kering
from keri.app import directing, indirecting, watching, habbing, storing
from keri.app.cli.common import existing

parser = argparse.ArgumentParser(description='Start watcher')
parser.set_defaults(handler=lambda args: startWatcher(args))
parser.add_argument('-H', '--http',
                    action='store',
                    default=5651,
                    help="Local port number the HTTP server listens on. Default is 5651.")
parser.add_argument('-T', '--tcp',
                    action='store',
                    default=5652,
                    help="Local port number the HTTP server listens on. Default is 5652.")
parser.add_argument('-n', '--name',
                    action='store',
                    default="watcher",
                    help="Name of controller. Default is watcher.")
parser.add_argument('-p', '--pre',
                    action='store',
                    default="",
                    help="Identifier prefix of controller of this watcher")



def startWatcher(args):
    name = args.name
    tcpPort = int(args.tcp)
    httpPort = int(args.http)

    help.ogler.level = logging.INFO
    help.ogler.reopen(name="keri", temp=True, clear=True)
    logger = help.ogler.getLogger()

    logger.info("\n******* Starting Watcher for %s listening: http/%s, tcp/%s "
                ".******\n\n", args.name, args.http, args.tcp)

    doers = setupWatcher(name, controller=args.pre, tcpPort=tcpPort, httpPort=httpPort)
    directing.runController(doers=doers, expire=0.0)

    logger.info("\n******* Ended Watcher for %s listening: http/%s, tcp/%s"
                ".******\n\n", args.name, args.http, args.tcp)


def setupWatcher(name="watcher", controller=None, tcpPort=5651, httpPort=5652):
    """
    """

    try:
        with habbing.existingHab(name=name, transferable=False) as hab:
            print("Watcher Identifier: {}".format(hab.pre))
            if hab.kever.prefixer.transferable:
                raise kering.ConfigurationError("watchers can only have a non-transferable identifier")
    except kering.ConfigurationError as e:
        print(f"identifier prefix for {name} does not exist, incept must be run first", )
        sys.exit(-1)

    hab, doers = existing.openHabitat(name=name, transferable=False)
    app = falcon.App(cors_enable=True)

    mbx = storing.Mailboxer(name=name)
    rep = storing.Respondant(hab=hab, mbx=mbx)

    kiwiServer = watching.KiwiServer(hab=hab, app=app, rep=rep, controller=controller)

    httpHandler = indirecting.HttpMessageHandler(hab=hab, app=app, rep=rep)
    mbxer = storing.MailboxServer(app=app, hab=hab, mbx=mbx)

    server = http.Server(port=httpPort, app=app)
    httpServerDoer = http.ServerDoer(server=server)

    server = serving.Server(host="", port=tcpPort)
    serverDoer = serving.ServerDoer(server=server)

    directant = directing.Directant(hab=hab, server=server)

    doers.extend([directant, serverDoer, mbxer, httpServerDoer, httpHandler, rep, kiwiServer])

    return doers



# -*- encoding: utf-8 -*-
"""
keri.kli.commands.multisig module

"""

import argparse

from hio import help
from hio.base import doing

from keri.app import directing, grouping, indirecting
from keri.app.cli.common import rotating, existing, displaying

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Begin or join a rotation of a group identifier')
parser.set_defaults(handler=lambda args: interactGroupIdentifier(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--group', '-g', help="Human readable environment reference for group identifier", required=True)
parser.add_argument('--data', '-d', help='Anchor data, \'@\' allowed', default=None, action="store", required=False)


def interactGroupIdentifier(args):
    """
    Performs an interaction event on the group identifier specified as an argument.  The identifier prefix of the
    environment represented by the name parameter must be a member of the group identifier.  This command will
    perform an interaction of the local identifier if the sequence number of the local identifier is the same as the
    group identifier sequence number.  It will wait for all other members of the group to acheive the same sequence
    number (group + 1) and then publish the signed interaction event for the group identifier to all witnesses and
    wait for receipts.

    Parameters:
        args (parseargs):  command line parameters

    """

    kwa = args.__dict__
    ixnDoer = GroupMultisigInteract(**kwa)

    doers = [ixnDoer]
    directing.runController(doers=doers, expire=0.0)


class GroupMultisigInteract(doing.DoDoer):
    """
    Command line DoDoer to launch the needed coroutines to run launch Multisig interaction.
       This DoDoer will remove the multisig coroutine and exit when it recieves a message
       that the multisig coroutine has successfully completed a cooperative rotation.

    """

    def __init__(self, name, **kwa):
        self.hab, doers = existing.openHabitat(name=name)
        self.rotr = grouping.MultiSigGroupDoer(hab=self.hab)
        self.msg = kwa

        mbd = indirecting.MailboxDirector(hab=self.hab, topics=['/receipt', '/multisig'])
        doers.extend([self.rotr, mbd])
        self.toRemove = list(doers)

        doers.extend([doing.doify(self.rotateDo)])

        super(GroupMultisigInteract, self).__init__(doers=doers)

    def rotateDo(self, tymth, tock=0.0, **opts):
        # enter context
        yield self.tock

        msg = dict(op=grouping.Ops.ixn, reason="Standard Interaction")
        msg["group"] = self.msg["group"]
        msg["data"] = self.msg["data"]

        self.rotr.msgs.append(msg)

        while not self.rotr.cues:
            yield self.tock

        rep = self.rotr.cues.popleft()
        print(rep)
        print()
        print("Group Identifier Rotation Complete:")
        displaying.printIdentifier(self.hab, rep["pre"])

        self.remove(self.toRemove)



import argparse
import logging

from keri.app import habbing, indirecting, directing
from keri import help

parser = argparse.ArgumentParser(description="Run a demo collection of witnesses")
parser.set_defaults(handler=lambda args: demo(args))

# help.ogler.level = logging.INFO
# logger = help.ogler.getLogger()


def demo(args):

    with habbing.openHab(name="wan", salt=b'wann-the-witness', transferable=False, temp=False) as wanHab, \
            habbing.openHab(name="wil", salt=b'will-the-witness', transferable=False, temp=False) as wilHab, \
            habbing.openHab(name="wes", salt=b'wess-the-witness', transferable=False, temp=False) as wesHab:

        wanDoers = indirecting.setupWitness(name="wan", hab=wanHab, temp=False, tcpPort=5632, httpPort=5642)
        wilDoers = indirecting.setupWitness(name="wil", hab=wilHab, temp=False, tcpPort=5633, httpPort=5643)
        wesDoers = indirecting.setupWitness(name="wes", hab=wesHab, temp=False, tcpPort=5634, httpPort=5644)

        doers = wanDoers + wilDoers + wesDoers
        directing.runController(doers, expire=0.0)



# -*- encoding: utf-8 -*-
"""
keri.peer.exchanging module

"""

from datetime import timedelta

from hio.base import doing
from hio.help import decking

from .. import help
from ..core import eventing, coring
from ..help import helping
from ..kering import ValidationError, MissingSignatureError, AuthZError

ExchangeMessageTimeWindow = timedelta(seconds=1010000)

logger = help.ogler.getLogger()


class Exchanger(doing.DoDoer):
    """
     Peer to Peer KERI message Exchanger.
    """

    def __init__(self, hab, handlers, controller=None, cues=None, delta=ExchangeMessageTimeWindow, **kwa):
        """
        Initialize instance

        Parameters:
            hab (Habitat): instance of local controller's
            handler(list): list of Handlers capable of responding to exn messages
            controller (str) qb64 prefix of the controlling identifier
            cues (Deck):  of Cues i.e. notices of requests needing response
            delta (timedelta): message timeout window
        """

        self.hab = hab
        self.controller = controller
        self.kevers = hab.kvy.kevers
        self.delta = delta
        self.routes = dict()
        self.cues = cues if cues is not None else decking.Deck()  # subclass of deque

        doers = []
        for handler in handlers:
            if handler.resource in self.routes:
                raise ValidationError("unable to register behavior {}, it has already been registered"
                                      "".format(handler.resource))

            self.routes[handler.resource] = handler
            doers.append(handler)

        super(Exchanger, self).__init__(doers=doers, **kwa)

    def processEvent(self, serder, source, sigers):
        """
        Process one serder event with attached indexed signatures representing
        a Peer to Peer exchange message.

        Parameters:
            serder (Serder) instance of event to process
            source (Prefixer) identifier prefix of event sender
            sigers (list) of Siger instances of attached controller indexed sigs

        """

        route = serder.ked["r"]
        payload = serder.ked["d"]
        dts = serder.ked["dt"]

        if route not in self.routes:
            raise AttributeError("unregistered route {} for exchange message = {}"
                                 "".format(route, serder.pretty()))

        behavior = self.routes[route]

        # delta = behavior.delta if behavior.delta is not None else self.delta
        delta = self.delta
        msgDt = helping.fromIso8601(dts)
        now = helping.nowUTC()

        if now - msgDt > delta:
            raise ValidationError("message received outside time window with delta {} message={}"
                                  "".format(delta, serder.pretty()))

        tholder, verfers = self.hab.verifiage(pre=source.qb64)

        if self.controller is not None and self.controller != source.qb64:
            raise AuthZError("Message {} is from invalid source {}"
                             "".format(payload, source.qb64))

        #  Verify provided sigers using verfers
        sigers, indices = eventing.verifySigs(serder=serder, sigers=sigers, verfers=verfers)
        if not tholder.satisfy(indices):  # at least one but not enough
            self.escrowPSEvent(serder=serder, sigers=sigers)
            raise MissingSignatureError("Failure satisfying sith = {} on sigs for {}"
                                        " for evt = {}.".format(tholder.sith,
                                                                [siger.qb64 for siger in sigers],
                                                                serder.ked))

        msg = dict(
            payload=payload,
            pre=source,
            sigers=sigers,
            verfers=verfers,
        )

        behavior.msgs.append(msg)

    def processResponseIter(self):
        """
        Iterate through cues and yields one or more responses for each cue.

        Parameters:
            cues is deque of cues

        """
        responses = []
        for _, behavior in self.routes.items():  # get responses from all behaviors
            while behavior.cues:
                cue = behavior.cues.popleft()
                responses.append(cue)

        while responses:  # iteratively process each response in responses
            msg = responses.pop(0)
            yield msg

    def escrowPSEvent(self, serder, sigers):
        """
        Escrow event that does not have enough signatures.

        Parameters:
            serder is Serder instance of event
            sigers is list of Siger instances of indexed controller sigs
        """
        pass


def exchange(route, payload, date=None, modifiers=None, version=coring.Version,
             kind=coring.Serials.json):
    """
    Create an `exn` message with the specified route and payload
    Parameters:
        route (string) to destination route of the message
        payload (dict) body of message to deliver to route
        date (str) Iso8601 formatted date string to use for this request
        modifiers (dict) equivalent of query string of uri, modifiers for the request that are not
                         part of the payload
        version (Version) is Version instance
        kind (Serials) is serialization kind

    """
    vs = coring.Versify(version=version, kind=kind, size=0)
    ilk = eventing.Ilks.exn
    dt = date if date is not None else helping.nowIso8601()

    ked = dict(v=vs,
               t=ilk,
               dt=dt,
               r=route,
               d=payload,
               q=modifiers
               )

    if modifiers is None:
        del ked["q"]

    return eventing.Serder(ked=ked)  # return serialized ked




# -*- encoding: utf-8 -*-
"""
keri.kli.commands.multisig module

"""

import argparse

from hio import help
from hio.base import doing

from keri.app import directing, grouping, indirecting
from keri.app.cli.common import rotating, existing, displaying

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Begin or join a rotation of a group identifier')
parser.set_defaults(handler=lambda args: rotateGroupIdentifier(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--group', '-g', help="Human readable environment reference for group identifier", required=True)

rotating.addRotationArgs(parser)


def rotateGroupIdentifier(args):
    """
    Performs a rotation on the group identifier specified as an argument.  The identifier prefix of the environment
    represented by the name parameter must be a member of the group identifier.  This command will perform a rotation
    of the local identifier if the sequence number of the local identifier is the same as the group identifier sequence
    number.  It will wait for all other members of the group to acheive the same sequence number (group + 1) and then
    publish the signed rotation event for the group identifier to all witnesses and wait for receipts.

    Parameters:
        args (parseargs):  command line parameters

    """

    kwa = args.__dict__
    rotDoer = GroupMultisigRotate(**kwa)

    doers = [rotDoer]
    directing.runController(doers=doers, expire=0.0)



class GroupMultisigRotate(doing.DoDoer):
    """
    Command line DoDoer to launch the needed coroutines to run launch Multisig rotation.
       This DoDoer will remove the multisig coroutine and exit when it recieves a message
       that the multisig coroutine has successfully completed a cooperative rotation.

    """

    def __init__(self, name, **kwa):
        self.hab, doers = existing.openHabitat(name=name)
        self.rotr = grouping.MultiSigGroupDoer(hab=self.hab)
        self.msg = kwa

        mbd = indirecting.MailboxDirector(hab=self.hab, topics=['/receipt', '/multisig'])
        doers.extend([self.rotr, mbd])
        self.toRemove = list(doers)

        doers.extend([doing.doify(self.rotateDo)])

        super(GroupMultisigRotate, self).__init__(doers=doers)


    def rotateDo(self, tymth, tock=0.0, **opts):
        # enter context
        yield self.tock

        msg = dict(op=grouping.Ops.rot, reason="Standard Rotation")
        msg["group"] = self.msg["group"]
        msg["sith"] = self.msg["sith"]
        msg["toad"] = self.msg["toad"]
        msg["data"] = self.msg["data"]
        msg["wits"] = self.msg["witnesses"] if "witnesses" in msg else []
        msg["cuts"] = self.msg["witness_cut"] if "witnesse_cut" in msg else []
        msg["adds"] = self.msg["witness_add"] if "witnesse_add" in msg else []


        self.rotr.msgs.append(msg)

        while not self.rotr.cues:
            yield self.tock

        rep = self.rotr.cues.popleft()
        print(rep)
        print()
        print("Group Identifier Rotation Complete:")
        displaying.printIdentifier(self.hab, rep["pre"])


        self.remove(self.toRemove)



import argparse
import json

from hio import help
from hio.base import doing

from keri import kering
from keri.app import directing, indirecting, grouping
from keri.app.cli.common import existing
from keri.peer import exchanging
from keri.vdr import issuing, verifying, viring

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Initialize a prefix')
parser.set_defaults(handler=lambda args: issueCredential(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--registry-name', '-r', help='Human readable name for registry, defaults to name of Habitat',
                    default=None)
parser.add_argument('--schema', '-s', help='qb64 SAID of Schema to issue',
                    default=None, required=True)
parser.add_argument('--type', '-t', help='Human readable type name of credential to issue',
                    default=None, required=True)
parser.add_argument('--source', '-c', help='AC/DC Source links',
                    default=None)
parser.add_argument('--recipient', '-R', help='qb64 identifier prefix of the recipient of the credential',
                    default=None)
parser.add_argument('--data', '-d', help='Credential data, \'@\' allowed', default=[], action="store", required=True)


def issueCredential(args):
    name = args.name
    if args.data is not None:
        try:
            if args.data.startswith("@"):
                f = open(args.data[1:], "r")
                data = json.load(f)
            else:
                data = json.loads(args.data)
        except json.JSONDecodeError:
            raise kering.ConfigurationError("data supplied must be value JSON to issue in a credential")
    else:
        raise kering.ConfigurationError("data supplied must be value JSON to issue in a credential")


    issueDoer = CredentialIssuer(name=name, registryName=args.registry_name, schema=args.schema, source=args.source,
                                 recipient=args.recipient, typ=args.type, data=data)

    doers = [issueDoer]
    directing.runController(doers=doers, expire=0.0)


class CredentialIssuer(doing.DoDoer):
    """
    Credential issuer DoDoer

    """

    def __init__(self, name, registryName, schema, source, recipient, typ, data):
        """

        Parameters:
             name:
             registryName:
             schema:
             source:
             recipient:
             typ: (str) human readable credential type name
             data: (dict) credential data dict
        """
        self.name = name
        self.hab, doers = existing.openHabitat(name=self.name)

        self.msg = dict(
            registryName=registryName,
            schema=schema,
            source=source,
            recipient=recipient,
            typ=typ,
            data=data
        )


        reger = viring.Registry(name=registryName)
        issuer = issuing.Issuer(hab=self.hab, name=registryName, reger=reger)
        self.verifier = verifying.Verifier(hab=self.hab, name=registryName, reger=reger, tevers=issuer.tevers)
        meh = grouping.MultisigEventHandler(hab=self.hab, verifier=self.verifier)

        handlers = [meh]
        exchanger = exchanging.Exchanger(hab=self.hab, handlers=handlers)

        mbx = indirecting.MailboxDirector(hab=self.hab, exc=exchanger, topics=["/receipt", "/multisig"])

        self.issr = issuing.IssuerDoer(hab=self.hab, issuer=issuer, verifier=self.verifier)
        doers.extend([self.issr, mbx, exchanger])
        self.toRemove = list(doers)

        doers.extend([doing.doify(self.issueDo)])
        super(CredentialIssuer, self).__init__(doers=doers)

    def issueDo(self, tymth, tock=0.0):
        """
        """
        yield self.tock

        self.issr.msgs.append(self.msg)

        done = False
        while not done:
            while self.verifier.cues:
                cue = self.verifier.cues.popleft()
                if cue["kin"] == "saved":
                    done = True

                yield self.tock
            yield

        self.remove(self.toRemove)



# -*- encoding: utf-8 -*-
"""
keri.vc.handling module

"""
from hio.base import doing
from hio.help import decking
from keri import kering
from keri.help import helping
from keri.vdr import issuing

from . import proving
from .. import help
from ..app import agenting
from ..core.coring import dumps, Deversify
from ..core.scheming import JSONSchema
from ..kering import ShortageError
from ..peer import exchanging
from ..vc.proving import Credentialer, buildProof

logger = help.ogler.getLogger()


class OfferHandler(doing.Doer):
    """
    Sample handler to a credential offer message from an Issuer to a Holder for a credential based

    on
        {
           "v": "KERI10JSON00011c_",                               // KERI Version String
           "t": "exn",                                             // peer to peer message ilk
           "dt": "2020-08-22T17:50:12.988921+00:00"
           "r": "/credential/offer"                                // resource with nested namespace
           "q": {
              "issuer": "did:keri:EEBp64Aw2rsjdJpAR0e2qCq3jX7q7gLld3LjAwZgaLXU"
              "output_descriptors: [
                 "EckOnHB11J4H9q16I3tN8DdpNXnCiP5QJQ7yvkWqTDdA"
              ],
              "format": {
                "cesr": {
                  "proof_type": ["Ed25519Signature2018"]
                }
              }
           } //embedded credential_manifest like structure, may contain presentation_definition of requirements for fullfilment
        }-AABAA1o61PgMhwhi89FES_vwYeSbbWnVuELV_jv7Yv6f5zNiOLnj1ZZa4MW2c6Z_vZDt55QUnLaiaikE-d_ApsFEgCA

    """

    resource = "/credential/offer"

    def __init__(self, wallet, formats, typ=JSONSchema(), cues=None, **kwa):
        """

        Parameters:
            wallet (Wallet) credential wallet that will hold the issued credentials
            formats (list) of format str names accepted for offers
            typ (JSONSchema) credential type to accept
        """
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()
        self.wallet = wallet
        self.formats = formats
        self.typ = typ

        super(OfferHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                payload = msg["payload"]
                recipient = msg["pre"]

                issuer = payload["issuer"]
                descriptors = payload["input_descriptors"]
                formats = payload["format"]

                schema = descriptors[0]

                fmts = []
                for fmt in self.formats:
                    if fmt in formats:
                        fmts.append(formats[fmt])

                if not fmts:
                    logger.info("No acceptable formats being offered in {}.  Needed one of {}."
                                "".format(formats, self.formats))
                    continue

                apply = credential_apply(issuer, schema, fmts)

                exn = exchanging.exchange(route="/credential/apply", payload=apply)
                self.cues.append(dict(dest=recipient, rep=exn))

                yield

            yield


class ApplyHandler(doing.DoDoer):
    """
        {
           "v": "KERI10JSON00011c_",                               // KERI Version String
           "t": "exn",                                             // peer to peer message ilk
           "dt": "2020-08-22T17:50:12.988921+00:00"
           "r": "/credential/apply"
           "q" {
              "issuer": "did:keri:EEBp64Aw2rsjdJpAR0e2qCq3jX7q7gLld3LjAwZgaLXU"
              "input_descriptors": [
                 "EckOnHB11J4H9q16I3tN8DdpNXnCiP5QJQ7yvkWqTDdA"
              ],
              "format": {
                 "cesr": {
                   "proof_type": ["Ed25519Signature2018"]
                 }
              }
           } //embedded credential_submission, may contain credential_fullfilment responding to presentation_def above
        }-AABAA1o61PgMhwhi89FES_vwYeSbbWnVuELV_jv7Yv6f5zNiOLnj1ZZa4MW2c6Z_vZDt55QUnLaiaikE-d_ApsFEgCA

    """

    resource = "/credential/apply"

    def __init__(self, hab, verifier, name, issuerCues=None, cues=None, **kwa):
        """

        Parameters:
            hab (Habitat) credential wallet that will hold the issued credentials
            verifier (Verifier) Local credential verifier used to verify and save any issued credential
        """
        self.hab = hab
        self.verifier = verifier
        self.name = name
        self.issuer = None
        self.issuerCues = issuerCues if issuerCues is not None else decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        self.msgs = decking.Deck()

        super(ApplyHandler, self).__init__(doers=[doing.doify(self.escrowDo)], **kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """
        yield self.tock
        self.issuer = issuing.Issuer(hab=self.hab, name=self.name, reger=self.verifier.reger, cues=self.issuerCues)

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                recipientIdentifier = msg["pre"]

                payload = msg["payload"]

                schema = payload["schema"]
                issuer = payload["issuer"]
                source = []

                if issuer != self.hab.pre:
                    logger.info("request for incorrect issuer {} to {}".format(issuer, self.hab.pre))
                    continue

                typ = payload["type"]
                if typ != "QualifiedvLEIIssuervLEICredential":
                    logger.info("credential type {} is invalid, only QualifiedvLEIIssuervLEICredential can be "
                                "auto-issued".format(typ))
                    continue


                types = ["VerifiableCredential", typ]


                data = payload["body"]
                dt = data["dt"] if "dt" in data else helping.nowIso8601()

                d = dict(
                    d="",
                    i=recipientIdentifier.qb64,
                    dt=dt,
                    t=types,
                )

                d |= data

                group = self.hab.group()
                if group is None:
                    pre = self.hab.pre
                else:
                    pre = group.gid

                creder = proving.credential(issuer=pre,
                                            schema=schema,
                                            subject=d,
                                            source=source,
                                            status=self.issuer.regk)
                try:
                    self.issuer.issue(creder=creder, dt=dt)
                except kering.MissingAnchorError:
                    logger.info("Missing anchor from credential issuance due to multisig identifier")

                craw = self.hab.endorse(creder)
                proving.parseCredential(ims=craw, verifier=self.verifier)

                yield self.tock

            yield self.tock

    def escrowDo(self, tymth, tock=0.0):
        """
        Returns:  doifiable Doist compatible generator method

        Usage:
            add result of doify on this method to doers list

        Processes the Groupy escrow for group icp, rot and ixn request messages.

        """
        # start enter context
        yield  # enter context
        while True:
            self.issuer.processEscrows()
            yield self.tock


class IssueHandler(doing.DoDoer):
    """
    Sample class that handles a credential Issue `exn` message.  By default, this handler
    verifies the credential with the provided verifier.  The incoming message must have the following format:

         {
       "vc" [
         {
           "vc": {
              "v": "KERI10JSON00011c_", //KERI Version String
              "x": "EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI", // Identifier prefix of the Schema
              "d": {
                   "type": [
                       "EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI"
                   ],
                   "id": "did:keri:EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI",
                   "issuer": "did:keri:EchZLZUFqtBGRWMh3Ur_iKucjsrFcxU7AjfCPko9CkEA",
                   "issuanceDate": "2021-06-09T17:35:54.169967+00:00",
                   "credentialSubject": {
                       "id": "did:keri:did:keri:Efaavv0oadfghasdfn443fhbyyr4v",
                       "lei": "254900OPPU84GM83MG36"
                   },
                   "credentialSchema": {
                       "id": ""
                       "type": ""
                   },
                   "credentialStatus": {
                      "id": "",
                      "type": ""
                   }
              }
           }, // embedded verifiable credential
           "proof": "-AABAA1o61PgMhwhi89FES_vwYeSbbWnVuELV_jv7Yv6f5zNiOLnj1ZZa4MW2c6Z_vZDt55QUnLaiaikE
                 -d_ApsFEgCA-GAB0AAAAAAAAAAAAAAAAAAAAABQEchZLZUFqtBGRWMh3Ur_iKucjsrFcxU7AjfCPko9CkEA"
           }
       ]   //list of verifiable credentials
    }


    """

    resource = "/credential/issue"

    def __init__(self, hab, verifier, cues=None, **kwa):
        """

        Parameters:
            wallet (Wallet) credential wallet that will hold the issued credentials
            typ (JSONSchema) credential type to accept
        """
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        self.verifier = verifier
        self.witq = agenting.WitnessInquisitor(hab=hab, klas=agenting.HttpWitnesser)

        doers = [self.witq, doing.doify(self.msgDo), doing.doify(self.verifierDo)]

        super(IssueHandler, self).__init__(doers=doers, **kwa)

    def msgDo(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifiying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                print("got message")
                payload = msg["payload"]
                envelopes = payload["vc"]
                for envlop in envelopes:
                    crd = envlop["vc"]
                    proof = envlop["proof"]

                    vs = crd["v"]

                    print("got credential")

                    kind, version, size = Deversify(vs)
                    raw = dumps(ked=crd, kind=kind)
                    if len(raw) != size:
                        raise ValueError("invalid length {} for credential = {}".format(size, crd))

                    msg = bytearray(raw)
                    msg.extend(proof.encode("utf-8"))

                    proving.parseCredential(ims=msg, verifier=self.verifier)
                yield

            yield

    def verifierDo(self, tymth, tock=0.0, **opts):
        """
        Process cues from Verifier coroutine

            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters
        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.verifier.cues:
                cue = self.verifier.cues.popleft()
                cueKin = cue["kin"]

                if cueKin == "saved":
                    creder = cue["creder"]

                    logger.info("Credential: %s, Schema: %s,  Saved", creder.said, creder.schema)
                    logger.info(creder.pretty())
                    print("Credential: {}, Schema: {},  Saved".format(creder.said, creder.schema))
                    print(creder.pretty())

                elif cueKin == "query":
                    qargs = cue["q"]
                    self.witq.query(**qargs)

                elif cueKin == "telquery":
                    qargs = cue["q"]
                    self.witq.telquery(**qargs)

                yield self.tock
            yield self.tock


class RequestHandler(doing.Doer):
    """
        Processor for a credential request with input descriptors in the payload used to
        match saved credentials based on a schema.  The payload of the request is expected to
        have the following format:

             {
                "submission_requirements": [{
                   "name": "Proof of LEI",
                   "rule": "pick",
                   "count": 1,
                   "from": "A"
                 }]
                "input_descriptors": [
                   {
                      "x":"EckOnHB11J4H9q16I3tN8DdpNXnCiP5QJQ7yvkWqTDdA",
                      "group": ["A"],
                   }
                ],
                "format": {
                   "cesr": {
                     "proof_type": ["Ed25519Signature2018"]
                   }
                }
             }

    """

    resource = "/presentation/request"

    def __init__(self, wallet, cues=None, **kwa):
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()
        self.wallet = wallet

        super(RequestHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """
        Process presentation request message with sender identifier, sigs and verfers

        Messages:
            payload is dict representing the body of a /presentation/request message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /presentation/request message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                payload = msg["payload"]
                requestor = msg["pre"]
                descriptors = payload["input_descriptors"]

                matches = []
                for descriptor in descriptors:
                    said = descriptor["x"]
                    credentials = self.wallet.getCredentials(said)
                    if len(credentials) > 0:
                        matches.append(credentials[0])

                if len(matches) > 0:
                    pe = presentation_exchange(matches)
                    exn = exchanging.exchange(route="/presentation/proof", payload=pe)
                    self.cues.append(dict(dest=requestor.qb64, rep=exn, topic="credential"))

                yield

            yield


class ProofHandler(doing.Doer):
    """
    Processor for responding to presentation proof peer to peer message.  The payload of the message
    is expected to have the following format:

        {
          "presentation_submission": {
              "id": "a30e3b91-fb77-4d22-95fa-871689c322e2",
              "definition_id": "32f54163-7166-48f1-93d8-ff217bdb0653",
              "descriptor_map": [
                   {
                     "id": "gleif_vlei",
                     "format": "cesr",
                     "path": "$.verifiableCredential[0]"
                   }
              ]
          },
          "verifiableCredential": [
              {
                "vc": {
                   "v": "KERI10JSON00011c_", //KERI Version String
                   "x": "EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI", // Identifier prefix of the Schema
                   "q": {
                        "type": [
                            "EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI"
                        ],
                        "id": "did:keri:EeyJ0eXBlIjogWyJWZXJpZmlhYmxlQ3JlZGVudGlhbCI",
                        "issuer": "did:keri:EchZLZUFqtBGRWMh3Ur_iKucjsrFcxU7AjfCPko9CkEA",  //Identifier prefix of
                        the issuer
                        "issuanceDate": "2021-06-09T17:35:54.169967+00:00",
                        "credentialSubject": {
                            "id": "did:keri:did:keri:Efaavv0oadfghasdfn443fhbyyr4v",
                            "lei": "254900OPPU84GM83MG36"
                        }
                    } // embedded verifiable credential
                }
                "proof": "-VA0-FABE4YPqsEOaPNaZxVIbY-Gx2bJgP-c7AH_K7pEE
                -YfcI9E0AAAAAAAAAAAAAAAAAAAAAAAElHzHwX3V6itsD2Ksg_CNBbUNTBYzLYw-AxDNI7_ZmaI
                -AABAALK_6pkUjCx76CTE7pNrKlhC84ewgZbs-4ciTvLcxkwUaWX7ukpr55P9RtJhlHxdtz3kN0zfM0HrtFitKuPy3BA"
              }
           ]
        }
    """

    resource = "/presentation/proof"

    def __init__(self, typ=JSONSchema(), cues=None, proofs=None, **kwa):
        self.msgs = decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()
        self.proofs = proofs if proofs is not None else decking.Deck()

        self.typ = typ

        super(ProofHandler, self).__init__(**kwa)

    def do(self, tymth, tock=0.0, **opts):
        """

        Handle incoming messages by parsing and verifying the credential and storing it in the wallet

        Parameters:
            payload is dict representing the body of a /credential/issue message
            pre is qb64 identifier prefix of sender
            sigers is list of Sigers representing the sigs on the /credential/issue message
            verfers is list of Verfers of the keys used to sign the message

        """

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                payload = msg["payload"]
                pre = msg["pre"]

                if "presentation_submission" not in payload:
                    raise ValueError("invalid presentation proof payload")

                if "verifiableCredential" not in payload:
                    raise ValueError("invalid presentation proof payload")

                pe = payload["presentation_submission"]
                vcs = payload["verifiableCredential"]

                if "descriptor_map" not in pe:
                    raise ValueError("invalud presentation submission in proof payload")

                dm = pe["descriptor_map"]

                for idx, descriptor in enumerate(dm):
                    # TODO:  Find verifiable credential in vcs based on `path`
                    vc = vcs[idx]
                    self.proofs.append((pre, vc))

                yield

            yield


def envelope(msg, typ=JSONSchema()):
    """
    Returns a dict of a VC split into the "vc" and "proof"

    Parameters:
        msg: bytes of verifiable credential to split
        typ: schema type of the VC
    """

    ims = bytearray(msg)
    try:
        creder = Credentialer(raw=ims)
    except ShortageError as e:
        raise e
    else:
        del ims[:creder.size]

    return dict(
        vc=creder.crd,
        proof=ims.decode("utf-8")
    )


def presentation_exchange(credentials):
    dm = []
    vcs = []

    for idx, (creder, prefixer, seqner, diger, sigers) in enumerate(credentials):
        proof = buildProof(prefixer, seqner, diger, sigers)
        dm.append(dict(
            id=creder.schema,
            format="cesr",
            path="$.verifiableCredential[{}]".format(idx)
        ))

        vcs.append(dict(
            vc=creder.crd,
            proof=proof.decode("utf-8")
        ))

    d = dict(
        presentation_submission=dict(
            descriptor_map=dm
        ),
        verifiableCredential=vcs
    )

    return d


def credential_apply(issuer, schema, typ, formats, body):
    """
        {
           "v": "KERI10JSON00011c_",                               // KERI Version String
           "t": "exn",                                             // peer to peer message ilk
           "dt": "2020-08-22T17:50:12.988921+00:00"
           "r": "/credential/apply"
           "q" {
              "issuer": "did:keri:EEBp64Aw2rsjdJpAR0e2qCq3jX7q7gLld3LjAwZgaLXU"
              "input_descriptors": [
                 "EckOnHB11J4H9q16I3tN8DdpNXnCiP5QJQ7yvkWqTDdA"
              ],
              "format": {
                 "cesr": {
                   "proof_type": ["Ed25519Signature2018"]
                 }
              },
              "body": {
                 // fields specific to the credential specified in the input_descriptor
              }
           } //embedded credential_submission, may contain credential_fullfilment responding to presentation_def above
        }

    Parameters:
        issuer (str) is qb64 identifier prefix of the issuer
        schema (str) is qb64 SAID of schema being applied for
        formats (list of CredentialFormat) is list of acceptable credential formats
        body (map) of values being applied for

    """

    d = dict(
        issuer=issuer,
        schema=schema,
        type=typ,
        body=body
    )

    for fmt in formats:
        d["format"].append(fmt.fmd)

    return d


def credential_issue(msgs, typ):
    """
    Returns a list of credentials enveloped inside a credential issue message

    Parameters:
        msgs (list) is list of CESR formatted, endorsed verifiable credentials
        typ (JSONSchema) type of credentials being issued

    """

    vcs = []
    for msg in msgs:
        vc = envelope(msg, typ)
        vcs.append(vc)

    pl = dict(
        vc=vcs
    )

    return pl



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse

from keri import kering
from keri.app import habbing

parser = argparse.ArgumentParser(description='Sign an arbitrary string')
parser.set_defaults(handler=lambda args: sign(args))
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--text', '-t', help='Text or file (starts with "@") to sign', required=True)


def sign(args):

    name = args.name

    try:
        with habbing.existingHab(name=name) as hab:

            txt = args.text
            if txt.startswith("@"):
                f = open(txt[1:], "r")
                data = f.read()
            else:
                data = txt

            sigers = hab.mgr.sign(ser=data.encode("utf-8"),
                                  verfers=hab.kever.verfers,
                                  indexed=True)

            for idx, siger in enumerate(sigers):
                print("{}. {}".format(idx+1, siger.qb64))

    except kering.ConfigurationError:
        print(f"prefix for {name} does not exist, incept must be run first", )
    except FileNotFoundError:
        print("unable to open file", args.text[1:])



# -*- encoding: utf-8 -*-
"""
keri.peer.httping module

"""
import json
from dataclasses import dataclass

import falcon
from hio.help import Hict

from keri import help
from keri import kering
from keri.core import coring
from keri.core.coring import Ilks
from keri.end import ending
from keri.help.helping import nowIso8601

logger = help.ogler.getLogger()

CESR_CONTENT_TYPE = "application/cesr+json"
CESR_ATTACHMENT_HEADER = "CESR-ATTACHMENT"
CESR_DATE_HEADER = "CESR-DATE"
CESR_RECIPIENT_HEADER = "CESR-RECIPIENT"


class SignatureValidationComponent(object):

    def __init__(self, hab, pre):
        self.hab = hab
        self.pre = pre

    def process_request(self, req, resp):
        sig = req.headers.get("SIGNATURE")

        ser = req.bounded_stream.read()
        if not self.validate(sig=sig, ser=ser):
            resp.complete = True
            resp.status = falcon.HTTP_401
            return
        req.context.raw = ser

    def validate(self, sig, ser):
        signages = ending.designature(sig)
        markers = signages[0].markers

        if self.pre not in self.hab.kevers:
            return False

        verfers = self.hab.kevers[self.pre].verfers
        for idx, verfer in enumerate(verfers):
            key = str(idx)
            if key not in markers:
                return False
            siger = markers[key]
            siger.verfer = verfer

            if not verfer.verify(siger.raw, ser):
                return False

        return True




@dataclass
class CesrRequest:
    resource: str
    date: str
    payload: dict
    modifiers: dict
    attachments: str


def parseCesrHttpRequest(req, prefix=None):
    """
    Parse Falcon HTTP request and create a CESR message from the body of the request and the two
    CESR HTTP headers (Date, Attachment).

    Parameters
        req (falcon.Request) http request object in CESR format:

    """
    if req.content_type != CESR_CONTENT_TYPE:
        raise falcon.HTTPError(falcon.HTTP_NOT_ACCEPTABLE,
                               title="Content type error",
                               description="Unacceptable content type.")

    try:
        data = json.load(req.bounded_stream)
    except ValueError:
        raise falcon.HTTPError(falcon.HTTP_400,
                               title="Malformed JSON",
                               description="Could not decode the request body. The "
                                           "JSON was incorrect.")

    resource = req.path
    if prefix is not None:
        resource = resource.removeprefix(prefix)

    if CESR_DATE_HEADER not in req.headers:
        raise falcon.HTTPError(falcon.HTTP_UNAUTHORIZED,
                               title="Date error",
                               description="Missing required date header.")

    dt = req.headers[CESR_DATE_HEADER]

    if CESR_ATTACHMENT_HEADER not in req.headers:
        raise falcon.HTTPError(falcon.HTTP_PRECONDITION_FAILED,
                               title="Attachment error",
                               description="Missing required attachment header.")
    attachment = req.headers[CESR_ATTACHMENT_HEADER]


    cr = CesrRequest(
        resource=resource,
        date=dt,
        payload=data,
        modifiers=req.params,
        attachments=attachment)

    return cr


def createCESRRequest(msg, client, date=None):
    """
    Turns a KERI message into a CESR http request against the provided hio http Client

    Parameters
       msg:  KERI message parsable as Serder.raw
       client: hio http Client that will send the message as a CESR request

    """

    dt = date if date is not None else nowIso8601()
    try:
        serder = coring.Serder(raw=msg)
    except kering.ShortageError as ex:  # need more bytes
        raise kering.ExtractionError("unable to extract a valid message to send as HTTP")
    else:  # extracted successfully
        del msg[:serder.size]  # strip off event from front of ims

    ilk = serder.ked["t"]
    attachments = bytearray(msg)
    query = serder.ked["q"] if "q" in serder.ked else None

    if ilk in (Ilks.icp, Ilks.rot, Ilks.ixn, Ilks.dip, Ilks.drt, Ilks.ksn, Ilks.rct):
        resource = "/kel"
        body = serder.raw
    elif ilk in (Ilks.req,):
        resource = "/" + ilk + "/" + serder.ked['r']
        body = serder.raw
    elif ilk in (Ilks.fwd,):
        resource = "/" + ilk + "/" + serder.ked['r']
        body = json.dumps(serder.ked["a"]).encode("utf-8")
    elif ilk in (Ilks.exn,):
        resource = "/" + ilk + serder.ked['r']
        body = json.dumps(serder.ked["d"]).encode("utf-8")
        dt = serder.ked["dt"]
    elif ilk in (Ilks.vcp, Ilks.vrt, Ilks.iss, Ilks.rev, Ilks.bis, Ilks.brv):
        resource = "/tel"
        body = serder.raw
    else:
        raise kering.InvalidEventTypeError("Event type {} is not handled by http clients".format(ilk))

    headers = Hict([
        ("Content-Type", CESR_CONTENT_TYPE),
        ("Content-Length", len(body)),
        (CESR_DATE_HEADER, dt),
        (CESR_ATTACHMENT_HEADER, attachments)
    ])

    client.request(
        method="POST",
        path=resource,
        qargs=query,
        headers=headers,
        body=body
    )


class InsecureSignatureComponent(object):

    def process_request(self, req, resp):
        ser = req.bounded_stream.read()
        req.context.raw = ser



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.witness module

Witness command line interface
"""
import argparse

import falcon
from hio.base import doing
from hio.core import http

parser = argparse.ArgumentParser()
parser.set_defaults(handler=lambda args: launch())


class Bootstrap:
    def on_post(self, req, rep):
        return falcon.HTTP_200


def launch():
    app = falcon.App()

    app.add_route('/bootstrap', Bootstrap())

    server = http.Server(port=5678, app=app)
    httpServerDoer = http.ServerDoer(server=server)

    doers = [httpServerDoer]
    tock = 0.03125
    doist = doing.Doist(limit=0.0, tock=tock, real=True)
    doist.do(doers=doers)


if __name__ == "__main__":
    print("launching bootstrap")
    launch()



from hio import help
from hio.base import doing
from hio.help import decking

from keri.app import agenting, grouping
from keri.vdr import issuing, viring

logger = help.ogler.getLogger()


class RegistryInceptDoer(doing.DoDoer):

    def __init__(self, hab, msgs=None, cues=None, **kwa):

        self.hab = hab
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()
        self.issuer = None
        self.gdoer = grouping.MultiSigGroupDoer(hab=hab)

        doers = [self.gdoer, doing.doify(self.inceptDo, **kwa)]
        super(RegistryInceptDoer, self).__init__(doers=doers)


    def inceptDo(self, tymth, tock=0.0, **kwa):
        """
        Returns:  doifiable Doist compatible generator method for creating a registry
        and sending its inception and anchoring events to witnesses or backers

        Usage:
            add result of doify on this method to doers list
        """
        # start enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)  # finish enter context

        while True:
            while self.msgs:
                msg = self.msgs.popleft()
                name = msg["name"]

                reger = viring.Registry(name=self.hab.name, temp=False)
                self.issuer = issuing.Issuer(hab=self.hab, name=name, reger=reger, noBackers=True, **kwa)
                self.extend([doing.doify(self.escrowDo), doing.doify(self.issuerDo)])
                yield self.tock

                while self.issuer.regk not in self.issuer.tevers:
                    yield self.tock

                yield self.tock

            yield self.tock


    def issuerDo(self, tymth, tock=0.0, **opts):
        """
        Process cues from credential issue coroutine

        Parameters:
            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value
            opts is dict of injected optional additional parameters
        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.issuer.cues:
                cue = self.issuer.cues.popleft()
                cueKin = cue['kin']
                if cueKin == "send":
                    tevt = cue["msg"]
                    witSender = agenting.WitnessPublisher(hab=self.hab, msg=tevt)
                    self.extend([witSender])

                    while not witSender.done:
                        _ = yield self.tock

                    self.remove([witSender])
                    self.cues.append(dict(kin="finished", regk=self.issuer.regk))

                elif cueKin == "kevt":
                    kevt = cue["msg"]
                    witDoer = agenting.WitnessReceiptor(hab=self.hab, msg=kevt)
                    self.extend([witDoer])

                    while not witDoer.done:
                        yield self.tock

                    self.remove([witDoer])

                elif cueKin == "multisig":
                    msg = dict(
                        op=cue["op"],
                        data=cue["data"],
                        reason=cue["reason"]
                    )
                    self.gdoer.msgs.append(msg)
                elif cueKin == "logEvent":
                    self.cues.append(dict(kin="finished", regk=self.issuer.regk))


                yield self.tock
            yield self.tock


    def escrowDo(self, tymth, tock=0.0):
        """
        Returns:  doifiable Doist compatible generator method

        Usage:
            add result of doify on this method to doers list

        Processes the Groupy escrow for group icp, rot and ixn request messages.

        """
        # start enter context
        yield  # enter context
        while True:
            self.issuer.processEscrows()
            yield



# -*- encoding: utf-8 -*-
"""
KERI
keri.app.agenting module

"""
import json

import falcon
from hio.base import doing
from hio.core.http import clienting
from hio.help import decking

from keri.app import keeping, obtaining, httping
from keri.core import coring, eventing, parsing
from .. import help, kering
from ..end import ending

logger = help.ogler.getLogger()


class KiwiServer(doing.DoDoer):
    """
    Routes for handling UI requests for Watcher Control

    """

    def __init__(self, hab, controller, cues=None, app=None, **kwa):
        self.hab = hab
        self.controller = controller
        self.app = app if app is not None else falcon.App(cors_enable=True)
        self.app.add_middleware(httping.SignatureValidationComponent(hab=self.hab, pre=self.controller))
        self.cues = cues if cues is not None else decking.Deck()

        self.app.add_route("/rotate", self, suffix="rotate")

        doers = []

        super(KiwiServer, self).__init__(doers=doers, **kwa)

    def on_post_rotate(self, req, rep):
        pre = self.controller
        prms = self.hab.ks.prms.get(self.hab.pre)

        aeid = self.hab.mgr.aeid
        cur = self.hab.kever

        algo = prms.algo
        salt = prms.salt
        tier = prms.tier
        pidx = prms.pidx

        ncount = 0  # next count
        code = coring.MtrDex.Ed25519N

        mgr = keeping.Manager(ks=self.hab.ks, aeid=aeid, pidx=pidx,
                              algo=algo, salt=salt, tier=tier)

        verfers, digers, cst, nst = mgr.incept(icount=1,
                                               ncount=ncount,
                                               isith=cur.tholder.sith,
                                               algo=keeping.Algos.randy,
                                               transferable=False,
                                               temp=False)


        opre = verfers[0].qb64  # old pre default move below to new pre from incept
        if digers:
            nxt = coring.Nexter(sith=nst,
                                digs=[diger.qb64 for diger in digers]).qb64
        else:
            nxt = ""

        serder = eventing.incept(keys=[verfer.qb64 for verfer in verfers],
                                 sith=cst,
                                 nxt=nxt,
                                 toad=cur.toad,
                                 wits=cur.wits,
                                 code=code)

        icpMsg = bytearray(serder.raw)
        sigers = mgr.sign(ser=serder.raw, verfers=verfers)
        icpMsg.extend(coring.Counter(code=coring.CtrDex.ControllerIdxSigs,
                                  count=len(sigers)).qb64b)  # attach cnt
        for sig in sigers:
            icpMsg.extend(sig.qb64b)  # attach sig


        sigers = self.hab.mgr.sign(ser=bytes(icpMsg),
                                   verfers=self.hab.kever.verfers,
                                   indexed=False)

        signage = ending.Signage(markers=sigers, indexed=False)
        sheaders = ending.signature([signage])

        self.hab.recreate(serder, opre, verfers)

        for key, val in sheaders.items():
            rep.append_header(key, val)

        rep.content_type = "application/json+CESR"
        rep.content_length = len(icpMsg)
        rep.data = icpMsg
        rep.status = falcon.HTTP_200


class WatcherClientRotateDoer(doing.DoDoer):
    """
    Sends KERI Auth'ed HTTP request to specified watcher to rotate identifer.  Current
    watcher set is then updated by removing the old watcher idenfitier and adding the
    new watcher identifier.

    """

    def __init__(self, hab, msgs=None, cues=None, **kwa):
        """
        Create doer to rotate remote watcher identifier prefix

        Parameters:
            name: is str name of Habitat
            watcher:
            kwa:
        """
        self.hab = hab
        self.msgs = msgs if msgs is not None else decking.Deck()
        self.cues = cues if cues is not None else decking.Deck()

        doers = [doing.doify(self.rotateDo)]
        super(WatcherClientRotateDoer, self).__init__(doers=doers, **kwa)


    def rotateDo(self, tymth, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list
        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            while self.msgs:
                watcher = self.msgs.popleft()
                habr = self.hab.db.habs.get(self.hab.name)
                if watcher not in habr.watchers:
                    raise kering.ValidationError("identifier {} is not a current watcher {}"
                                                 "".format(watcher, habr.watchers))


                payload = dict(pre=self.hab.pre)
                raw = json.dumps(payload)
                sigers = self.hab.mgr.sign(ser=raw.encode("utf-8"),
                                           verfers=self.hab.kever.verfers,
                                           indexed=True)

                signage = ending.Signage(markers=sigers, indexed=True)
                headers = ending.signature([signage])

                loc = obtaining.getwitnessbyprefix(watcher)
                client = clienting.Client(hostname=loc.ip4, port=loc.http)
                clientDoer = clienting.ClientDoer(client=client)
                self.extend([clientDoer])

                client.request(method="POST", path="/rotate", headers=headers, body=raw)
                while not client.responses:
                    yield self.tock

                resp = client.respond()
                if resp.status != 200:
                    print("Invalid status from watcher:", type(resp.status))
                    return

                if not self.authenticate(watcher=watcher, resp=resp):
                    print("Invalid response from watcher")
                    return

                wat = self.processWatcherResponse(watcher=watcher, icp=bytes(resp.body))

                self.remove([clientDoer])

                self.cues.append(dict(old=watcher, new=wat))

                yield self.tock

            yield self.tock


    def processWatcherResponse(self, watcher, icp):
        ctrlKvy = eventing.Kevery(db=self.hab.db)
        parsing.Parser().parse(ims=bytearray(icp), kvy=ctrlKvy)

        srdr = coring.Serder(raw=bytearray(icp))
        wat = srdr.pre

        habr = self.hab.db.habs.get(self.hab.name)
        ewats = set(habr.watchers)

        ewats.remove(watcher)
        ewats.add(wat)

        habr.watchers = list(ewats)

        self.hab.db.habs.pin(self.hab.name, habr)
        return wat


    @staticmethod
    def authenticate(watcher, resp):
        if "Signature" not in resp.headers:
            return False

        signages = ending.designature(resp.headers["Signature"])

        cigar = signages[0].markers[watcher]
        verfer = coring.Verfer(qb64=watcher)
        if not verfer.verify(cigar.raw, bytes(resp.body)):
            return False

        return True
